/*******************************************************************************
 * SemDiff
 * Copyright (C) 2008 Barthelemy Dagenais
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this library. If not, see 
 * <http://www.gnu.org/licenses/lgpl-3.0.txt>
 *******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.view;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.*;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.eclipse.ui.texteditor.ITextEditor;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.jface.action.*;
import org.eclipse.jface.commands.ActionHandler;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.*;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.SWT;
import org.eclipse.compare.CompareConfiguration;
import org.eclipse.compare.CompareEditorInput;
import org.eclipse.compare.CompareUI;
import ca.mcgill.cs.swevo.semdiff.core.CoreActivator;
import ca.mcgill.cs.swevo.semdiff.core.CoreListener;
import ca.mcgill.cs.swevo.semdiff.core.DetectorResults;
import ca.mcgill.cs.swevo.semdiff.core.Profile;
import ca.mcgill.cs.swevo.semdiff.core.Repository;
import ca.mcgill.cs.swevo.semdiff.core.Result;
import ca.mcgill.cs.swevo.semdiff.core.SemDiffFacade;
import ca.mcgill.cs.swevo.semdiff.core.Transaction;
import ca.mcgill.cs.swevo.semdiff.ui.UIActivator;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;				
import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.CastConverter;
import ca.mcgill.cs.swevo.diffcat.view.actions.AssignSeedAction;
import ca.mcgill.cs.swevo.diffcat.view.actions.ClearAllSeedsAction;
import ca.mcgill.cs.swevo.diffcat.view.actions.CompareAction;
import ca.mcgill.cs.swevo.diffcat.view.actions.GotoTransactionAction;
import ca.mcgill.cs.swevo.diffcat.view.actions.NextTransactionAction;
import ca.mcgill.cs.swevo.diffcat.view.actions.PreviousTransactionAction;
import ca.mcgill.cs.swevo.diffcat.view.actions.PrintBenchmarkEntryAction;
import ca.mcgill.cs.swevo.diffcat.view.diff.FileCompareEditorInput;
import ca.mcgill.cs.swevo.diffcat.view.filters.DiffFilterChainBuilder;

/**
 * This file is a modified version of SemDiff's CategoryView.  
 */
public class CategoryView extends ViewPart implements CoreListener, ISelectionChangedListener, IOpenListener, IPropertyChangeListener
//, IClusterManagerListener	TODO 
{

	/**
	 * The ID of the view as specified by the extension.
	 */
	public static final String ID = "ca.mcgill.cs.swevo.diffcat.view.CategoryView";

	private Action doubleClickAction;
	
	private CategoryViewContentProvider contentProvider;

	private final String J_EDITOR_ID = PlatformUI.getWorkbench().getEditorRegistry().getDefaultEditor(".java").getId();
	
	class NameSorter extends ViewerSorter {
	}

	/**
	 * The constructor.
	 */
	public CategoryView() {
		super ();
		coreActivator = CoreActivator.getDefault();
		facade = coreActivator.getFacade();
	}

	private TreeViewer transactionViewer;

	private Label transactionLabel;

	private Action collapseAllAction;

	private Action clearAction;

	private Action previousTransactionAction;

	private Action nextTransactionAction;

	private Action compareAction;

	private AssignSeedAction assignSeedAction;
	
	private Action removeAllSeedsAction;
	
	private Action printBenchmarkEntryAction;
	
	private GotoTransactionAction gotoTransaction;

	private Text usernameText;

	private Text dateText;

	private Text commitText;

	private Transaction currentTransaction;

	private String[] shortDescriptions;

	private final SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

	private CoreActivator coreActivator;

	private SemDiffFacade facade;

	private Repository currentRepository;

	private TreeViewNavigator navigator;
	
	private Set<IFilter<DiffCatResult>> adHocFilters = new HashSet<IFilter<DiffCatResult>>();
	
	private DiffFilterChainBuilder filterChainBuilder = new DiffFilterChainBuilder();
	
	public void changeTransaction(int index) {
		Transaction previousTransaction = currentTransaction;

		// Precheck
		if (currentRepository == null || index < 0
				|| index >= currentRepository.getTransactionsSize()) {
			MessageDialog
					.openError(
							getSite().getShell(),
							"Invalid Transaction ID",
							"The transaction id you entered is invalid. It must be between 0 and the maximum number of transactions.");
			return;
		}

		// In case something goes wrong.
		try {
			currentTransaction = facade.getTransaction(currentRepository, index);
		} catch (Exception e) {
			MessageDialog
					.openError(getSite().getShell(), "Transaction could not be loaded",
							"The transaction " + index
									+ " could not be loaded because this error occurred: "
									+ e.getMessage());
		}

		if (currentTransaction == null) {
			MessageDialog
					.openWarning(
							getSite().getShell(),
							"Transaction not initialized",
							"The transaction "
									+ index
									+ " is not initialized. Hint: initialize the transactions, fetch files or run detectors to initialize it.");
			// Restore
			currentTransaction = previousTransaction;
		} else {
			clearFilters();
			refreshDetails(currentTransaction);
		}
	}

	private void clear() {
		transactionViewer.setInput(null);
		transactionLabel.setText("");
		usernameText.setText("");
		dateText.setText("");
		commitText.setText("");
		gotoTransaction.setEnabled(false);
		nextTransactionAction.setEnabled(false);
		previousTransactionAction.setEnabled(false);
	}

	private void createActions() {
		collapseAllAction = new Action() {
			@Override
			public void run() {
				transactionViewer.collapseAll();
			}
		};
		collapseAllAction.setImageDescriptor(AbstractUIPlugin.imageDescriptorFromPlugin(
				UIActivator.PLUGIN_ID, "icons/collapseall.gif"));
		collapseAllAction.setToolTipText("Collapse all");

		clearAction = new Action() {
			@Override
			public void run() {
				clear();
			}
		};
		clearAction.setImageDescriptor(AbstractUIPlugin.imageDescriptorFromPlugin(UIActivator.PLUGIN_ID,
				"icons/removeall.gif"));
		clearAction.setToolTipText("Clear view");

		gotoTransaction = new GotoTransactionAction(this);
		gotoTransaction.setToolTipText("Goto transaction");
		gotoTransaction.setImageDescriptor(AbstractUIPlugin.imageDescriptorFromPlugin(
				UIActivator.PLUGIN_ID, "icons/goto_obj.gif"));
		gotoTransaction.setEnabled(false);

		previousTransactionAction = new PreviousTransactionAction(this);
		previousTransactionAction.setEnabled(false);
		IHandlerService handler = (IHandlerService) getSite().getService(IHandlerService.class);
		handler.activateHandler("ca.mcgill.cs.swevo.diffcat.view.actions.previousTransaction",
				new ActionHandler(previousTransactionAction));

		nextTransactionAction = new NextTransactionAction(this);
		nextTransactionAction.setEnabled(false);
		handler = (IHandlerService) getSite().getService(IHandlerService.class);
		handler.activateHandler("ca.mcgill.cs.swevo.diffcat.view.actions.nextTransaction", new ActionHandler(
				nextTransactionAction));
		
		compareAction = new CompareAction(this);
		compareAction.setDescription("Compare files");
		compareAction.setText("Compare files");
		
		assignSeedAction = new AssignSeedAction(this);
		assignSeedAction.setDescription("Assign To Seed");
		assignSeedAction.setText("Assign to seed");
		
		removeAllSeedsAction = new ClearAllSeedsAction(this);
		removeAllSeedsAction.setDescription("Clear all seeds");
		removeAllSeedsAction.setText("Remove all seeds");
		
		printBenchmarkEntryAction = new PrintBenchmarkEntryAction(this);
		printBenchmarkEntryAction.setDescription("Print benchmark entry");
		printBenchmarkEntryAction.setText("Print benchmark entry");
		
		doubleClickAction = new CompareAction(this);
	}

	
	
	private void createCommit(Composite parent) {
		Composite commitContainer = new Composite(parent, SWT.NONE);
		GridLayout gLayout = new GridLayout(2, false);
		gLayout.marginHeight = 0;
		gLayout.marginWidth = 0;
		commitContainer.setLayout(gLayout);
		GridData gData = new GridData(GridData.FILL_HORIZONTAL);
		gData.verticalAlignment = SWT.TOP;
		commitContainer.setLayoutData(gData);

		Label commitLabel = new Label(commitContainer, SWT.NONE);
		commitLabel.setText("Commit: ");
		gData = new GridData();
		gData.verticalAlignment = SWT.TOP;
		commitLabel.setLayoutData(gData);

		commitText = new Text(commitContainer, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL);
		commitText.setEditable(false);
		gData = new GridData(GridData.FILL_HORIZONTAL);
		gData.heightHint = 50;
		gData.verticalAlignment = SWT.TOP;
		commitText.setLayoutData(gData);
	}

	private void createDetails(Composite parent) {
		Composite detailsContainer = new Composite(parent, SWT.NONE);
		GridLayout gLayout = new GridLayout(2, false);
		gLayout.marginHeight = 0;
		gLayout.marginWidth = 0;
		detailsContainer.setLayout(gLayout);
		GridData gData = new GridData();
		gData.verticalAlignment = SWT.TOP;
		detailsContainer.setLayoutData(gData);

		Label userLabel = new Label(detailsContainer, SWT.NONE);
		userLabel.setText("Username: ");
		gData = new GridData();
		gData.verticalAlignment = SWT.TOP;
		userLabel.setLayoutData(gData);

		usernameText = new Text(detailsContainer, SWT.BORDER);
		usernameText.setEditable(false);
		gData = new GridData();
		gData.widthHint = 125;
		gData.verticalAlignment = SWT.TOP;
		usernameText.setLayoutData(gData);

		Label dateLabel = new Label(detailsContainer, SWT.NONE);
		dateLabel.setText("Date: ");
		gData = new GridData();
		gData.verticalAlignment = SWT.TOP;
		dateLabel.setLayoutData(gData);

		dateText = new Text(detailsContainer, SWT.BORDER);
		dateText.setEditable(false);
		gData = new GridData();
		gData.widthHint = 125;
		gData.verticalAlignment = SWT.TOP;
		dateText.setLayoutData(gData);
	}

	private void createHeader(Composite parent) {
		Composite headerComposite = new Composite(parent, SWT.BORDER);

		GridLayout gLayout = new GridLayout(3, false);
		gLayout.horizontalSpacing = 15;
		headerComposite.setLayout(gLayout);

		GridData gData = new GridData(GridData.FILL_HORIZONTAL);
		headerComposite.setLayoutData(gData);

		createLabel(headerComposite);
		createDetails(headerComposite);
		createCommit(headerComposite);
	}

	private void createLabel(Composite parent) {
		transactionLabel = new Label(parent, SWT.NONE);
		GridData gData = new GridData();
		gData.widthHint = 75;
		transactionLabel.setLayoutData(gData);
	}

	private Composite container = null;
	
	@Override
	public void createPartControl(Composite parent) {
		
		navigator = new TreeViewNavigator();
		currentRepository = coreActivator.getCurrentRepository();
		
		container = new Composite(parent, SWT.NONE);
		container.setLayout(new GridLayout(1, false));

		createHeader(container);
		createViewer(container);
		createActions();
		createToolbar();
		hookDoubleClickAction();
		hookContextMenu();
		contributeToActionBars();
		
		if (currentRepository != null) {
			newRepository(currentRepository);
		}
		
		coreActivator.addCoreListener(this);
//		ClusterManager.getManager().addListener(this);	TODO, don't need this
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
	}
	
	private void fillLocalPullDown(IMenuManager menu) {
		hookSeedContextMenu();
//		// those in no seed
		menu.add(new Separator());
	}
	
	public void createToolbar() {
		IActionBars bars = getViewSite().getActionBars();
		bars.getToolBarManager().add(clearAction);
		bars.getToolBarManager().add(gotoTransaction);
		bars.getToolBarManager().add(previousTransactionAction);
		bars.getToolBarManager().add(nextTransactionAction);
		
	}

	private void createViewer(Composite parent) {
		
		Tree tree = new Tree(parent, SWT.MULTI | SWT.V_SCROLL | SWT.H_SCROLL);
		GridData gData = new GridData(GridData.FILL_BOTH);
		tree.setLayoutData(gData);
		
		transactionViewer = new TreeViewer(tree);
		contentProvider = new CategoryViewContentProvider (this);
		transactionViewer.setContentProvider(contentProvider);
		transactionViewer.setLabelProvider(new CategoryViewLabelProvider());
		transactionViewer.addSelectionChangedListener(this);
		transactionViewer.setSorter(new NodeSorter());
		transactionViewer.addOpenListener(this);
		
		Activator.getDefault().getPreferenceStore().addPropertyChangeListener(this);
	}

	public Collection<DiffCatResult> getVisibleDiffs () {
//		
		Tree tree = transactionViewer.getTree();
		TreeItem[] items = tree.getItems();
		
		Collection<DiffCatResult> vis = new LinkedList<DiffCatResult>();
		getVisibleDiffs(items, vis);
		
		return vis;
	}

	private void getVisibleDiffs(TreeItem[] items, Collection<DiffCatResult> vis) {
		for (TreeItem item : items) 
		{
			Object data = item.getData();
			if (data instanceof DiffCatNode) {
				vis.add(((DiffCatNode) data).getResult());
			}
			else {
				getVisibleDiffs (item.getItems(), vis);
			}
		}
	}
	
	@Override
	public void dispose() {
		coreActivator.removeCoreListener(this);
		super.dispose();
	}

	public Repository getCurrentRepository() {
		return currentRepository;
	}

	public Transaction getCurrentTransaction() {
		return currentTransaction;
	}

	public String[] getShortDescriptions() {
		return shortDescriptions;
	}

	public void newRepository(Repository repository) {
		if (repository.getTransactionsSize() > 0) {
			currentTransaction = facade.getTransaction(repository, 0);
			if (currentTransaction != null) {
				int max = repository.getTransactionsSize() - 1;
				transactionLabel.setText("0 of " + max);
				gotoTransaction.setEnabled(true);
				nextTransactionAction.setEnabled(true);
				previousTransactionAction.setEnabled(true);
				clearFilters();
				refreshDetails(currentTransaction);
			}
		}
	}

	@Override
	public void open(OpenEvent event) {

	}

	@Override
	public void profileChanged(Profile oldProfile, Profile newProfile) {
		if (newProfile == null
				|| (oldProfile != null && !oldProfile.getName().equals(newProfile.getName()))) {
			currentRepository = null;
			currentTransaction = null;
			clear();
		}
	}

	public ISelection getTreeSelection () {
		return transactionViewer.getSelection();
	}
	
	public void highlight (Set<DiffCatResult> results) {
		List<DiffCatNode> nodes = new ArrayList<DiffCatNode>(results.size());
		for (DiffCatResult result : results) {
			nodes.add(new DiffCatNode(result));
		}
		IStructuredSelection sel = new StructuredSelection(nodes);
		transactionViewer.expandAll();
		transactionViewer.setSelection(sel);
	}
	
	public void refreshDetails(Transaction transaction) {
		int max = currentRepository.getTransactionsSize() - 1;
		int id = transaction.getId();
		String username = currentTransaction.getUserName();
		Date aDate = currentTransaction.getDate();
		String date = "";
		if (aDate != null) {
			date = sdf.format(currentTransaction.getDate());
		} else { }
		String commit = currentTransaction.getCommitComment();

		transactionLabel.setText(id + " of " + max);
		transactionLabel.redraw();
		usernameText.setText(username != null ? username : "");
		usernameText.redraw();
		dateText.setText(date != null ? date : "");
		dateText.redraw();
		commitText.setText(commit != null ? commit : "");
		commitText.redraw();
		transactionViewer.setInput(transaction);
		transactionViewer.expandAll();
	}

	private void clearFilters () {
		this.adHocFilters.clear();
		ClusterManager.getManager().clearData();
	}
	
	@Override
	public void repositoryChanged(Repository oldRepository, Repository newRepository) {
		if (newRepository != null) {
			currentTransaction = null;
			currentRepository = newRepository;
			newRepository(newRepository);
		}
	}

	@Override
	public void selectionChanged(SelectionChangedEvent event) {
	}

	private void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("#PopupMenu");
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			@Override
			public void menuAboutToShow(IMenuManager manager) {
				CategoryView.this.fillContextMenu(manager);
			}
		});
		Menu menu = menuMgr.createContextMenu(transactionViewer.getControl());
		transactionViewer.getControl().setMenu(menu);
		getSite().registerContextMenu(menuMgr, transactionViewer);
	}

	private void hookSeedContextMenu () {
		IMenuManager menuManager = getViewSite().getActionBars().getMenuManager();
		menuManager.setRemoveAllWhenShown(true);
		menuManager.addMenuListener(new IMenuListener() {
			@Override
			public void menuAboutToShow(IMenuManager manager) {
				CategoryView.this.fillSeedDisplayMenu(manager);
			}
		});
	}
		
	private void fillContextMenu(IMenuManager manager) {
		
		final Object obj = getSelection();
		
		if (obj instanceof DiffCatNode) {
							
			manager.add(removeAllSeedsAction);
			manager.add(printBenchmarkEntryAction);
			
			manager.add(new Separator());
			
			// shows the diff in the editor
			manager.add(compareAction);
			
			// assigns a seed to the diff
			manager.add(assignSeedAction);				
			
			// lets user remove seeds from diff
			DiffCatNode node = (DiffCatNode) obj;		//TODO: add back in later
			final DiffCatResult result = node.getResult();
			final ClusterManager manager2 = ClusterManager.getManager();
			Set<String> seeds = manager2.getSeedsFor(result);
			
			if (! seeds.isEmpty()) {
				manager.add(new Separator());
			}
			
			for (final String seed : seeds) {
				manager.add(new Action() {
					@Override
					public String getText () { 
						return "Remove from " + seed;
					}
					@Override
					public void run () {
						assignSeedAction.removeFromSeed(seed, result);
						HasSeedFilter h = new HasSeedFilter(seed);
						if (CollectionUtil.contains(adHocFilters, h)) {
							// check whether the filter is even still relevant 
							// (i.e., whether there's anything under it)
							if (! manager2.seedExists(seed)) {
								SetUtil.removeAll(adHocFilters, h);
							}
							refreshDetails(currentTransaction);
						}
					}
				});
			}
			
			// allow easy addition to existing seeds
			if (!manager2.getSeeds().isEmpty()) {
				manager.add(new Separator());
			}
			Set<String> allSeeds = manager2.getSeeds();
			
			for (final String seed : allSeeds) {
				if (! seeds.contains(seed)) {
					manager.add(new Action() {
						@Override
						public String getText () {
							return "Assign to " + seed;
						}
						@Override
						public void run () {
							assignSeedAction.assignSeed(seed, result);
						}
					});
				}
			}
		}
		else if (obj instanceof MethodSigNode || obj instanceof ClassSigNode || obj instanceof FieldSigNode) {
			manager.add(removeAllSeedsAction);
			manager.add(assignSeedAction);			
			manager.add(printBenchmarkEntryAction);
			manager.add(new Separator());
			manager.add(compareAction);
			
			TreeParent parent = (TreeParent) obj;
			final String signature = parent.getName();
			
			final ClusterManager manager2 = ClusterManager.getManager();
			Set<String> elementSeeds = manager2.getSeedsFor(signature);
			
			if (!elementSeeds.isEmpty()) {
				manager.add(new Separator());
			}
			
			// these are seeds we can remove it from
			for (final String seed : elementSeeds)
			{
				manager.add(new Action() {
					@Override
					public String getText () { 
						return "Remove from " + seed;
					}
					@Override
					public void run () {
						assignSeedAction.removeFromSeed(seed, obj);
						HasSeedFilter h = new HasSeedFilter(seed);
						if (CollectionUtil.contains(adHocFilters, h)) {
							// check whether the filter is even still relevant 
							// (i.e., whether there's anything under it)
							if (! manager2.seedExists(seed)) {
								SetUtil.removeAll(adHocFilters, h);
							}
							refreshDetails(currentTransaction);
						}
					}
				});
			}
			
			if (!manager2.getSeeds().isEmpty()) {
				manager.add(new Separator());
			}
			Set<String> allSeeds = manager2.getSeeds();
			
			for (final String seed : allSeeds) {
				if (! elementSeeds.contains(seed)) {
					manager.add(new Action() {
						@Override
						public String getText () {
							return "Assign to " + seed;
						}
						@Override
						public void run () {
							assignSeedAction.assignSeed(seed, obj);
						}
					});
				}
			}
		}
		
	}

	private void fillSeedDisplayMenu (IMenuManager menu) {
		final ClusterManager manager = ClusterManager.getManager();			//
		
		// build filters for all seeds that currently exist
		for (final String seed : manager.getSeeds()) {
			final HasSeedFilter filter = new HasSeedFilter(seed);
			// if we've already applied the filter, then we're going to remove it now
			if (CollectionUtil.contains(adHocFilters, filter)) {
				menu.add(new Action() {
					@Override
					public String getText () { return "Remove " + seed + " filter";}
					@Override
					public void run () { 
						SetUtil.removeAll(adHocFilters, filter);
						refreshDetails(currentTransaction);
					}
				});
			}
			// otherwise we're going to apply it now
			else {
				menu.add(new Action() {
					@Override
					public String getText () { return "Apply " + seed + " filter";}
					@Override
					public void run () {
						adHocFilters.add(new SeedFilter(seed));
						refreshDetails(currentTransaction);
					}
				});
			}
		}
		
		// build the "no seed" filter
		// if already has no seed, then we can only remove it
		if (CollectionUtil.contains(adHocFilters, NoSeedFilter.class)) {
			menu.add(new Action() {
				@Override
				public String getText () { return "Hide no-seed filter"; }
				@Override
				public void run () {
					SetUtil.removeAll(adHocFilters, NoSeedFilter.class);
					refreshDetails(currentTransaction);
				}
			});
		} 
		// otherwise we can add it
		else {
			menu.add(new Action() {
				@Override
				public String getText () { return "Apply no-seed filter"; }
				@Override
				public void run () {
					SetUtil.removeAll(adHocFilters, SeedFilter.class);
					adHocFilters.add(new NoSeedFilter());
					refreshDetails(currentTransaction);
				}
			});
		}

		// all 
		menu.add(new Action() {
			@Override
			public void run() {
				removeSeedFilters();
				refreshDetails(currentTransaction);
			}
			@Override
			public String getText () {
				return "Remove all filters";
			}
		});
	}
	
	public void removeSeedFilters () {
		SetUtil.removeAll(adHocFilters, SeedFilter.class);
		SetUtil.removeAll(adHocFilters, NoSeedFilter.class);
	}
	
	private class SeedFilter implements IFilter<DiffCatResult> 
	{
		private final String seed;
		public SeedFilter (String seed) { this.seed = seed; }
		@Override
		public boolean passesFilter(DiffCatResult t) {
			return ClusterManager.getManager().isPartOfSeed(seed, t);
		}
		public String getSeed () { return seed; }
	}
	
	private class NoSeedFilter implements IFilter<DiffCatResult> {
		@Override
		public boolean passesFilter(DiffCatResult t) {
			return ! ClusterManager.getManager().isPartOfSeed(t);
		}
	}
	
	private class HasSeedFilter implements IFilter<IFilter<DiffCatResult>> {
		private final String seed;
		public HasSeedFilter (String seed) {this.seed = seed;}
		@Override
		public boolean passesFilter(IFilter<DiffCatResult> t) {
			if (t instanceof SeedFilter) {
				return ((SeedFilter) t).getSeed().equals(seed);
			} else {
				return false;
			}
		}
	}
	
	private Object getSelection () {
		ISelection selection = transactionViewer.getSelection();
		return ((IStructuredSelection)selection).getFirstElement();
	}
	
	private void hookDoubleClickAction() {
		transactionViewer.addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				doubleClickAction.run();
			}
		});
	}
	
	/**
	 * Passing the focus request to the viewer's control.
	 */
	@Override
	public void setFocus() {
		transactionViewer.getControl().setFocus();
	}
	
	public void compareNodeInEditor (ClassSigNode node) {
		_compareNodeInEditor(node);
	}
	
	public void compareNodeInEditor (MethodSigNode node) {
		_compareNodeInEditor(node);
	}
	
	public void compareNodeInEditor (FieldSigNode node) {
		_compareNodeInEditor(node);
	}
	
	private void _compareNodeInEditor (TreeParent node) {
		
		String leftPath = contentProvider.getLeftPath(node);
		String rightPath = contentProvider.getRightPath(node);
		
		if (leftPath == null || rightPath == null) {
			return;
		}
		
		LRTuple<LRTuple<Integer>> vals = contentProvider.getPositionsOf(node);
		LRTuple<Integer> froms = vals.getLeft();
		LRTuple<Integer> tos = vals.getRight();
		
		DiffCatResult shell = new DiffCatResult();
		shell.setPathFrom(leftPath);
		shell.setPathTo(rightPath);
		shell.setChangeType(ChangeOperation.METHOD_INVOCATION_UPDATE.name());
		shell.setPosFrom(froms.getLeft());
		shell.setLenFrom(froms.getRight());
		shell.setPosTo(tos.getLeft());
		shell.setLenTo(tos.getRight());
		
		compareDiffInEditor(shell);
	}
	
	public void compareDiffInEditor (DiffCatResult result) {
		
		if (result.getChangeType().equals(ChangeOperation.CLASS_DELETE.name()))
		{
			ITextEditor editor = openEditor(result.getPathFrom());
			editor.selectAndReveal(result.getPosFrom(), result.getLenFrom());
			return;
		}
		
		if (result.getChangeType().equals(ChangeOperation.CLASS_INSERT.name())) 
		{
			ITextEditor editor = openEditor(result.getPathTo());
			editor.selectAndReveal(result.getPosTo(), result.getLenTo());
			return;
		}
		
		if (result.getChangeType().equals(ChangeOperation.FUNCTIONALITY_DELETE.name()) || 
			result.getChangeType().equals(ChangeOperation.ATTRIBUTE_DELETE.name()) || 
			result.getChangeType().equals(ChangeOperation.TRIVIAL_CONSTRUCTOR_DELETE.name())) 
		{
			if (result.getPathTo() == null) { // i.e., belongs to class delete
				ITextEditor editor = openEditor(result.getPathFrom());
				editor.selectAndReveal(result.getPosFrom(), result.getLenFrom());
				return;
			}
		}
		
		if (result.getChangeType().equals(ChangeOperation.FUNCTIONALITY_INSERT.name()) ||
			result.getChangeType().equals(ChangeOperation.ATTRIBUTE_INSERT.name()) || 
			result.getChangeType().equals(ChangeOperation.TRIVIAL_CONSTRUCTOR_INSERT.name())) 
		{
			if (result.getPathFrom() == null) {	// i.e., belongs to class insert
				ITextEditor editor = openEditor(result.getPathTo());
				editor.selectAndReveal(result.getPosTo(), result.getLenTo());
				return;
			}
		}
		
		// open the editor
		CompareEditorInput input = getDefaultCompareEditorInput(result);
		CompareUI.openCompareEditor(input);
		
		// navigate to the right diff
		navigator.navigateToDiff(input, result);
	}
	
	private CompareEditorInput getDefaultCompareEditorInput(DiffCatResult result) {
		
		String pathFrom = result.getPathFrom();
		String pathTo   = result.getPathTo();
		
		Repository repository = coreActivator.getCurrentRepository();
		
		FileCompareEditorInput input = null;

		File file1 = facade.getSourceFile(repository, pathFrom);
		File file2 = facade.getSourceFile(repository, pathTo);
		if (file1 != null && file2 != null) {
			CompareConfiguration configuration = new CompareConfiguration();
			input = new FileCompareEditorInput(configuration, file1, file2, result);
		}
		
		return input;
	}
	
	private ITextEditor openEditor (String path) 
	{
		try {
			Repository repository = coreActivator.getCurrentRepository();
			File file = facade.getSourceFile(repository, path);
			IWorkbenchPage page = getViewSite().getPage();
			
			return (ITextEditor) IDE.openEditor(page, file.toURI(), J_EDITOR_ID, false);
		} catch (PartInitException e) {
			return null;
		} 
	}

	@Override
	public void propertyChange(PropertyChangeEvent event) {
		if (filterChainBuilder.isFilterProperty(event.getProperty())) {
			transactionViewer.refresh();
		}
	}
	
	public Set<IFilter<DiffCatResult>> getAdHocFilters () {
		return adHocFilters;
	}

//	@Override
//	public void notifySeedChange() {
//		// have to refresh the seed view! 
//		refreshDetails(currentTransaction);
//	}
}