/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.cd;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.evolizer.changedistiller.distilling.Distiller;
import org.evolizer.changedistiller.model.entities.ClassHistory;
import org.evolizer.changedistiller.model.entities.SourceCodeChange;

import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsFieldResultFilter;
import ca.mcgill.cs.swevo.util.SetUtil;

/**
 * Maps SemDiffJavaSource units to input handled by the ChangeDistiller
 * component   
 * 
 * @author david
 */
public class ChangeDistillerProxy {

	private static ChangeDistillerProxy proxy;
	
	static {
		try {
			proxy = new ChangeDistillerProxy();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static ChangeDistillerProxy getProxy () {
		return proxy;
	}
	
	private final Logger log = Logger.getLogger(getClass().getName());
	
	/**
	 * Used to compare two arbitrary nodes
	 */
	private final String HACK_METHOD_NAME = "hackMethod";
	
	/**
	 * Used to compare two methods 
	 */
	private final String HACK_CLASS_NAME = "hackCLASS";
	
	/**
	 * Stuff required to use {@link changeDistiller}. 
	 */
	private final String PROJECT_NAME = "__TMP_PROJECT";
	private final String LEFT_FILE_NAME = "__LEFT_FILE.java";
	private final String RIGHT_FILE_NAME = "__RIGHT_FILE.java";
	private final IProject project;

	/**
	 * The actual implementation from Fluri et. al
	 */
	private final Distiller changeDistiller;
	
	/**
	 * Used to map CD diffs to DiffCat diffs
	 */
	private final CDDiffMapper mapper;
	
	private ChangeDistillerProxy () throws AdaptorException {
		project = makeProject();
		changeDistiller = new Distiller();
		mapper = new CDDiffMapper();
	}
	
	/**
	 * Creates a temporary project
	 * @return Reference to the created project
	 * @throws AdaptorException On any error
	 */
	private IProject makeProject () throws AdaptorException {
		
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		IProject project = root.getProject(PROJECT_NAME);
		try {
			if (project.exists()) {
				return project;
			}
			
			project.create(null);
			project.open(null);
			
			return project; 
			
		} catch (Exception e) {
			throw new AdaptorException(e);
		}
	}
	
	/**
	 * Distills all ChangeDistiller diffs between the two nodes. 
	 * The two nodes must be of the same kind (i.e., class or method)
	 */
	public Set<DiffCatResult> distill (ASTNode left, ASTNode right)
	{
		if (left.getNodeType() != right.getNodeType()) {
			throw new IllegalArgumentException("Nodes must be of the same type.");
		}

		if (left instanceof CompilationUnit) {
			return distillNormal(left, right);
		}
		
		if (left instanceof AbstractTypeDeclaration) {
			return distillTypeDecls((AbstractTypeDeclaration) left, (AbstractTypeDeclaration) right);
		}
		
		if (left instanceof MethodDeclaration) {
			return distillMethodDecls ((MethodDeclaration) left, (MethodDeclaration) right);
		}
		
		return distillOtherNodes(left, right);
	}
	
	public Set<DiffCatResult> distillMethods (MethodDeclaration left, String leftChunk, MethodDeclaration right, String rightChunk)
	{
		String leftContent = mapToClass(HACK_CLASS_NAME, leftChunk);
		String rightContent = mapToClass(HACK_CLASS_NAME, rightChunk);
		
		try {
			IFile leftFile = mapToFile (leftContent, LEFT_FILE_NAME);
			IFile rightFile = mapToFile(rightContent, RIGHT_FILE_NAME);
			
			return processFiles (left, leftFile, right, rightFile);
		}
		catch (AdaptorException e) {
			log.log(Level.SEVERE, "Error. Did not retrieve results.");
			return new HashSet<DiffCatResult>();
		}
	}
	
	public Set<DiffCatResult> distillCodeChunks (ASTNode left, String leftChunk, ASTNode right, String rightChunk)
	{
		String leftContent = mapToMethodInClass(leftChunk);
		String rightContent = mapToMethodInClass(rightChunk);
		
		try {
			IFile leftFile = mapToFile (leftContent, LEFT_FILE_NAME);
			IFile rightFile = mapToFile(rightContent, RIGHT_FILE_NAME);
			
			return processFiles (left, leftFile, right, rightFile);
		}
		catch (AdaptorException e) {
			log.log(Level.SEVERE, "Error. Did not retrieve results.");
			return new HashSet<DiffCatResult>();
		}
	}
	
	private Set<DiffCatResult> distillTypeDecls (AbstractTypeDeclaration left, AbstractTypeDeclaration right) {
		
		String leftName = left.getName().getIdentifier();
		String rightName = right.getName().getIdentifier();
		
		if (leftName.equals(rightName)) {
			return distillNormal(left, right);
		}
		else {
			return distillTypeDeclsDifferentName(left, right);
		}
	}
	
	/**
	 * left >= AbstractTypeDeclaration
	 * right >= AbstractTypeDeclaration
	 */
	private Set<DiffCatResult> distillNormal (ASTNode left, ASTNode right){
		
		try {
			IFile leftFile = mapToFile(left, LEFT_FILE_NAME);
			IFile rightFile = mapToFile(right, RIGHT_FILE_NAME);
			
			Set<DiffCatResult> results = processFiles(left, leftFile, right, rightFile); 
			results = SetUtil.removeAll(results, new IsFieldResultFilter());
			
			if (left instanceof AbstractTypeDeclaration) {
				doExtraDiffs((AbstractTypeDeclaration) left, (AbstractTypeDeclaration) right, results);
			}
			// left instanceof CompilationUnit
			else {
				doExtraDiffs ((CompilationUnit) left, (CompilationUnit) right, results);
			}
			
			return results;
		} 
		catch (AdaptorException e) 
		{
			log.log(Level.SEVERE, "Error. Did not retrieve results.");
			return new HashSet<DiffCatResult>();
		}
	}

	private void doExtraDiffs(AbstractTypeDeclaration left, AbstractTypeDeclaration right, Set<DiffCatResult> results) {
		FieldOpDetector detector = FieldOpDetector.getDetector();
		results.addAll (detector.getAttributeOps(left, right));
		
		ThrownExceptionsDetector thrownDetector = ThrownExceptionsDetector.getDetector();
		results.addAll (thrownDetector.getThrownExceptionUpdates(left, right, results));
	}
	
	private void doExtraDiffs (CompilationUnit left, CompilationUnit right, Set<DiffCatResult> knowns)
	{
		List<AbstractTypeDeclaration> leftDecls = ASTUtil.getTypeDeclarations(left);
		List<AbstractTypeDeclaration> rightDecls = ASTUtil.getTypeDeclarations(right);
		
		for (int i = 0 ; i < leftDecls.size(); i++) {
			AbstractTypeDeclaration leftDecl = leftDecls.get(i);
			String leftName = leftDecl.getName().getIdentifier();
			for (int j = 0; j < rightDecls.size(); j++) {
				AbstractTypeDeclaration rightDecl = rightDecls.get(j);
				String rightName = rightDecl.getName().getIdentifier();
				
				if (leftName.equals(rightName)) {
					doExtraDiffs(leftDecl, rightDecl, knowns);
				}
			}
		}
	}
	
	private Set<DiffCatResult> distillTypeDeclsDifferentName (AbstractTypeDeclaration left, AbstractTypeDeclaration right) {
		
		String rightName = right.getName().getIdentifier();
		String leftName = left.getName().getIdentifier();
		
		// change the class and constructor names to match .. else CD won't line them up
		adaptNames(right, rightName, leftName);
		
		Set<DiffCatResult> results = distillNormal (left, right);
		
		// revert the class and constructor names to their original
		adaptNames(right, leftName, rightName);
		
		return results;
	}
	
	private Set<DiffCatResult> distillOtherNodes (ASTNode left, ASTNode right) {
		
		String leftContent = mapToMethodInClass (left); 
		String rightContent = mapToMethodInClass (right);
		
		try {
			IFile leftFile = mapToFile (leftContent, LEFT_FILE_NAME);
			IFile rightFile = mapToFile(rightContent, RIGHT_FILE_NAME);
			
			return processFiles (left, leftFile, right, rightFile);
		}
		catch (AdaptorException e) {
			log.log(Level.SEVERE, "Error. Did not retrieve results.");
			return new HashSet<DiffCatResult>();
		}
	}

	private Set<DiffCatResult> distillMethodDecls (MethodDeclaration left, MethodDeclaration right) {
		
		String leftContent = mapToClass(HACK_CLASS_NAME, left);
		String rightContent = mapToClass(HACK_CLASS_NAME, right);
		
		try {
			IFile leftFile = mapToFile (leftContent, LEFT_FILE_NAME);
			IFile rightFile = mapToFile(rightContent, RIGHT_FILE_NAME);
			
			return processFiles (left, leftFile, right, rightFile);
		}
		catch (AdaptorException e) {
			log.log(Level.SEVERE, "Error. Did not retrieve results.");
			return new HashSet<DiffCatResult>();
		}
	}
	
	private Set<DiffCatResult> processFiles (ASTNode node1, IFile file1, ASTNode node2, IFile file2)
	{
		// distill
		changeDistiller.performDistilling(file1, file2);
		
		// retrieve
		Collection<SourceCodeChange> sourceCodeChanges = changeDistiller.getSourceCodeChanges();
		
		// map
		mapper.setLeftParent(node1);
		mapper.setRightParent(node2);
		if (sourceCodeChanges == null) {
			return new HashSet<DiffCatResult>();
		}
		Set<DiffCatResult> results = mapper.mapCDResults(sourceCodeChanges);
		mapper.setLeftParent(null);
		mapper.setRightParent(null);
		mapper.getClassRenames().clear();
		mapper.getFieldSignatureUpdates().clear();
		mapper.getMethodSignatureUpdates().clear();
		
		// manage previous changes
		if (changeDistiller.getSourceCodeChanges() != null) {
			changeDistiller.getSourceCodeChanges().clear();
		}
		
		// some memory management (don't know if necessary)
		if (changeDistiller.getClassHistory() != null) {
			ClassHistory classHistory = changeDistiller.getClassHistory();
			classHistory.getAttributeHistories().clear();
			classHistory.getInnerClassHistories().clear();
			classHistory.getMethodHistories().clear();
			classHistory.getVersions().clear();
		}
		
		return results;
	}
	
	private String mapToClass (String className, ASTNode node)  {
		return mapToClass(className, node.toString());
	}
	
	private String mapToClass (String className, String node) 
	{
		StringBuffer buf = new StringBuffer();
		buf.append("public class ");
		buf.append(className);
		buf.append(" {\n");
		buf.append(node);
		buf.append("\n}");
		
		return buf.toString();
	}
	
	private String mapToMethodInClass(ASTNode node) {
		return mapToMethodInClass (node.toString());
	}
	
	private String mapToMethodInClass(String content) {
		
		StringBuffer buf = new StringBuffer();
		buf.append("public class ");
		buf.append(HACK_CLASS_NAME);
		buf.append(" {\n");
		buf.append("void ");
		buf.append(HACK_METHOD_NAME);
		buf.append("() {\n");
		buf.append(content);
		buf.append("\n}\n}");
		
		return buf.toString();
	}
	
	private IFile mapToFile (ASTNode declaration, String fileName) throws AdaptorException
	{
		return mapToFile (declaration.toString(), fileName);
	}
	
	private IFile mapToFile (String content, String fileName) throws AdaptorException
	{
		IFile file = project.getFile(fileName);
		
		try {
			if (file.exists()) {
				file.delete(true, null);
			}
			
			file.create(new ByteArrayInputStream(content.getBytes()), true, null);
		} catch (CoreException e ) {
			throw new AdaptorException(e);
		}
		
		return file;
	}
	
	private void adaptNames(AbstractTypeDeclaration right, String oldName, String newName) {
		right.getName().setIdentifier(newName);
		right.accept(new ConstructorNameChanger(oldName, newName));
	}
}
