/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.cd.finder;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.evolizer.changedistiller.model.classifiers.ChangeType;
import org.evolizer.changedistiller.model.entities.Delete;
import org.evolizer.changedistiller.model.entities.Insert;
import org.evolizer.changedistiller.model.entities.Move;
import org.evolizer.changedistiller.model.entities.SourceCodeChange;
import org.evolizer.changedistiller.model.entities.SourceCodeEntity;
import org.evolizer.changedistiller.model.entities.Update;

import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.util.ArrayUtil;
import ca.mcgill.cs.swevo.util.LRTuple;


/**
 * Test this guy on Spring Transaction #33, 189
 * @author david
 *
 */
public class UltraMethodDeclarationFinder extends ANodeFinder {

	private String oldName;
	private final String[] newParamNames;
	private final String[][] newParamTypes;
	private String[] declType;
	
	private Collection<SourceCodeChange> changes;
	
	private final Logger log;
	
	private Map<String, String> renamed;
	
	List<MethodDeclaration> secondaries = new LinkedList<MethodDeclaration>();
	
	public UltraMethodDeclarationFinder (MethodDeclaration right, Collection<SourceCodeChange> changes)
	{
		initOldName (right, changes);
		this.newParamNames = ASTUtil.getParamNames(right);
		this.newParamTypes = ASTUtil.getParamTypesViaStrings(right);
		this.changes = convertReorderings(changes);
		this.renamed = new HashMap<String, String>();
		this.log = Logger.getLogger(getClass().getName());
		if (right.resolveBinding() != null) {
			this.declType = BindingUtil.getDeclaringType(right.resolveBinding());
		} else {
			this.declType = BindingUtil.UNKNOWN_TYPE_BINDING;
		}
		initRenameMap(changes);
	}
	
	private Collection<SourceCodeChange> convertReorderings (Collection<SourceCodeChange> changes) {
		Collection<SourceCodeChange> result = new LinkedList<SourceCodeChange>();
		for (SourceCodeChange change : changes) {
			if (! change.getChangeType().equals(ChangeType.PARAMETER_ORDERING_CHANGE)) {
				result.add(change);
				continue;
			}
			Move move = (Move) change;
			LRTuple<SourceCodeChange> pair = convertToPair (move);
			result.add(pair.getLeft());
			result.add(pair.getRight());
		}
		return result;
	}
	
	private LRTuple<SourceCodeChange> convertToPair (Move move)
	{
		SourceCodeChange left = new Delete(ChangeType.PARAMETER_DELETE, move.getRootEntity(), move.getChangedEntity(), move.getParentEntity());
		SourceCodeChange right= new Insert(ChangeType.PARAMETER_INSERT, move.getRootEntity(), move.getChangedEntity(), move.getParentEntity());
		return new LRTuple<SourceCodeChange>(left, right);
	}
	
	private void initOldName (MethodDeclaration right, Collection<SourceCodeChange> changes)
	{
		SourceCodeChange methodRename = null;
		for (SourceCodeChange change : changes) {
			if (change.getChangeType().equals(ChangeType.METHOD_RENAMING)) {
				methodRename = change;
			}
		}
		
		if (methodRename == null) {
			oldName = right.getName().toString();
		}
		else {
			oldName = methodRename.getChangedEntity().getUniqueName();
		}
	}
	
	private void initRenameMap (Collection<SourceCodeChange> changes) {
		for (SourceCodeChange change : changes) {
			if (change.getChangeType().equals(ChangeType.PARAMETER_RENAMING)) {
				String oldName = CDUtil.getParamName(change.getChangedEntity());
				String newName = CDUtil.getParamName(((Update) change).getNewEntity());
				renamed.put(oldName, newName);
			}
		}
	}
	
	public boolean visit (MethodDeclaration declaration)
	{
		final class BooleanWrapper {
			public Boolean bool = true;
		}
		
		String oldName = declaration.getName().toString();
		
		if (! this.oldName.equals(oldName)) {
			return true;
		}
		
		String[][] arrayParamTypes = ASTUtil.getParamTypesViaStrings(declaration);
		List<String[]> paramTypes = new ArrayList<String[]>();
		for (int i = 0; i < arrayParamTypes.length; i++){
			paramTypes.add(arrayParamTypes[i]);
		}
		
		String[] arrayParamNames = ASTUtil.getParamNames(declaration);
		List<String> paramNames = new ArrayList<String>();
		for (int i = 0; i < arrayParamNames.length; i++) {
			paramNames.add(arrayParamNames[i]);
		}
		
		// not true if one or more renamed params is missing
		if (! applyParamRenames (paramNames) ) {
			return true;
		}
		
		// not true if one or more deleted params is missing
		if (! applyParamDeletes(paramNames, paramTypes)) {
			return true;
		}
		
		// 
		final BooleanWrapper foundAll = new BooleanWrapper();
		List<Insert> inserts = retrieveInserts ();
		Collections.sort(inserts, new Comparator<Insert>() {
			@Override
			public int compare(Insert arg0, Insert arg1) {
				String argName0 = CDUtil.getParamName(arg0.getChangedEntity());
				String argName1 = CDUtil.getParamName(arg1.getChangedEntity());
				
				int i0 = ArrayUtil.indexOf(newParamNames, argName0);
				int i1 = ArrayUtil.indexOf(newParamNames, argName1);
				
				if (i0 == -1 || i1 == -1) {
					foundAll.bool = false;
				}
				
				return i0 - i1;
			}
		});
		
		// did not find all inserts
		if (! foundAll.bool) {
			return true;
		}
		
		for (int j = 0; j < inserts.size(); j++) {
			Insert insert = inserts.get(j);
			String paramName = CDUtil.getParamName(insert.getChangedEntity());
			int index = ArrayUtil.indexOf(newParamNames, paramName);
			if (index == -1) {
				return false;
			}
			String[] paramType = newParamTypes[index];
			if (index == 0) {
				applyParamInsert(paramNames, paramTypes, paramName, paramType, 0);
			}
			else {
				String leftParamName = newParamNames[index-1]; // inserted beside this guy
				int i = paramNames.indexOf(leftParamName);
				if (i == -1) {
					// didn't find something (should not happen)
					return true;
				}
				applyParamInsert(paramNames, paramTypes, paramName, paramType, i+1);
			}
		}
		
		// not true if one or more param types is missing
		if (! applyTypeUpdates (paramNames, paramTypes)) {
			return true;
		}
		
		// now that have applied all ops, check if method decl stats match!
		if (! ArrayUtil.areEqual(newParamNames, paramNames)) {
			return true;
		}
		
		if (newParamTypes.length != paramTypes.size()) {
			return true;
		}
		
		for (int i = 0; i < newParamTypes.length; i++) {
			String[] newParamType = newParamTypes[i];
			String[] nowParamType = paramTypes.get(i);
			String s1 = newParamType[newParamType.length - 1];
			String s2 = nowParamType[nowParamType.length - 1];
			if (! s1.equals(s2)) {
				if (s1.endsWith("[]")) {
					s1 = s1.substring(0, s1.length() - 2);
				}
				if (s2.endsWith("[]")) {
					s2 = s2.substring(0, s2.length() - 2);
				}
				
				if (s1.equals(s2)) {
					log.warning("Warning! Param type equality by ignoring array brackets. CD bug.");
					continue;
				}
				else {
					return true;
				}
			}
		}
		
		// if we got here, there is potential
		return setAssocNode (declaration);
	} 
	
	private boolean setAssocNode(MethodDeclaration node) {
		if (super.assocNode == null) {
			super.assocNode = node;
			return false;
		}
		else {
			String declType[] = BindingUtil.getDeclaringType(((MethodDeclaration) super.assocNode).resolveBinding());
			if (super.fullyMatches(declType, this.declType)) {
				return false;
			}
			
			String declType2[] = BindingUtil.getDeclaringType(node.resolveBinding());
			if (super.fullyMatches(declType2, this.declType)) {
				super.assocNode = node;
				return false;
			}
			
			if (! super.fullyOrPartialMatches(declType, this.declType) && 
				super.fullyOrPartialMatches(declType2, this.declType)) {
				super.assocNode = node;
				return false;
			}
			
			if (! super.fullyOrPartialMatches(declType2, this.declType)) {
				return false;
			}
			
			// otherwise can only store as secondary (figure out if we need to)
			secondaries.add(node);
			return false;
		}
	}
	
	private List<Insert> retrieveInserts () {
		ArrayList<Insert> inserts = new ArrayList<Insert>();
		for (SourceCodeChange change : changes) {
			if (change.getChangeType().equals(ChangeType.PARAMETER_INSERT)) {
				inserts.add((Insert) change);
			}
		}
		return inserts;
	}
	
	private boolean applyTypeUpdates(List<String> paramNames, List<String[]> paramTypes) {
		boolean appliedAll = true;
		for (SourceCodeChange change : changes) {
			if (change.getChangeType().equals(ChangeType.PARAMETER_TYPE_CHANGE)) {
				appliedAll &= applyTypeUpdate (paramNames, paramTypes, (Update) change);
			}
		}
		return appliedAll;
	}
	
	private boolean applyTypeUpdate (List<String> paramNames, List<String[]> paramTypes, Update update)
	{
		String name = CDUtil.getParamName(update.getChangedEntity());
		
		int index = -1;
		if (renamed.containsKey(name)) {
			index = paramNames.indexOf(renamed.get(name));
		}
		else {
			index = paramNames.indexOf(name);
		}
		
		if (index == -1) {
			return false;
		}
		
		paramTypes.set(index, CDUtil.getParamType(update.getNewEntity()));
		return true;
	}
	
	private boolean applyParamDeletes (List<String> paramNames, List<String[]> paramTypes)
	{
		boolean appliedAll = true;
		for (SourceCodeChange change : changes) {
			if (change.getChangeType().equals(ChangeType.PARAMETER_DELETE)) {
				appliedAll &= applyParamDelete (paramNames, paramTypes, (Delete) change);
			}
		}
		
		return appliedAll;
	}
	
	private boolean applyParamDelete(List<String> paramNames, List<String[]> paramTypes, Delete change) {
		
		String deleted = CDUtil.getParamName(change.getChangedEntity());
		if (! paramNames.contains(deleted)) {
			return false;
		}
		int indexOf = paramNames.indexOf(deleted);
		paramNames.remove(indexOf);
		paramTypes.remove(indexOf);
		
		return true;
	}

	private boolean applyParamRenames (List<String> paramNames)
	{
		boolean appliedAll = true;
		for (SourceCodeChange change : changes) {
			if (change.getChangeType().equals(ChangeType.PARAMETER_RENAMING)) {
				appliedAll &= applyParamRename (paramNames, (Update) change);
			}
		}
		return appliedAll;
	}
	
	private void applyParamInsert (List<String> paramNames, List<String[]> paramTypes, String name, String[] type, int where)
	{
		paramNames.add(where, name);
		paramTypes.add(where, type);
	}
	
	private boolean applyParamRename (List<String> paramNames, Update rename) 
	{
		SourceCodeEntity changedEntity = rename.getChangedEntity();
		String uniqueName = CDUtil.getParamName(changedEntity);
		for (int i = 0; i < paramNames.size(); i++) {
			if (paramNames.get(i).equals(uniqueName)) {
				paramNames.set(i, CDUtil.getParamName(rename.getNewEntity()));
				return true;
			}
		}
		
		return false;
	}
}
