/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.model;

import java.util.List;

import javax.persistence.Basic;
import javax.persistence.DiscriminatorValue;
import javax.persistence.Entity;
import javax.persistence.Transient;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.hibernate.annotations.Type;

import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.semdiff.core.Result;

/**
 * Main data class. Represents inferred structural diffs.
 * To cluster some common diffs:
 * @see DiffCodes 
 * 
 * @author david
 */

@Entity
@DiscriminatorValue(value = "DiffCatResult")
public class DiffCatResult extends Result {

	private final static String TAG = "@@";
	
	private String changeType;
	
	/**
	 * The ASTNodes containing the diff 
	 * (don't represent the diff itself)
	 */
	private ASTNode leftParent;
	private ASTNode rightParent;
	
	/**
	 * Location in source code of diff ASTNodes.
	 * Used by DiffCatView
	 */
	private int posFrom;
	private int lenFrom;
	private int posTo;
	private int lenTo;
	
	private String leftMethodSignature;
	private String rightMethodSignature;
	private String leftClassSignature;
	private String rightClassSignature;
	private String leftFieldSignature;
	private String rightFieldSignature;
	
	public DiffCatResult () {
		this.posFrom = this.posTo = 0;
		this.lenFrom = this.lenTo = -1;
	}
	
	public DiffCatResult (Result r) {
		this.handleFrom = r.getHandleFrom();
		this.handleTo = r.getHandleTo();
		this.pathFrom = r.getPathFrom();
		this.pathTo = r.getPathTo();
		setNodeFrom (r.getNodeFrom());
		setNodeTo   (r.getNodeTo());
		this.transactionId = r.getTransactionId();
		this.repositoryId = r.getRepositoryId();
	}
	
	/**
	 * This ASTNode represents the parent of the left node (i.e., its enclosing method, field, or class)
	 * It is only available before persistence. 
	 */
	@Transient
	public ASTNode getLeftParent() {
		return leftParent;
	}

	/**
	 * This ASTNode represents the parent of the right node (i.e., its enclosing method, field, or class)
	 * It is only available before persistence. 
	 */
	@Transient
	public ASTNode getRightParent() {
		return rightParent;
	}

	/**
	 * Used by Hibernate when loading persisted diffs. Clients shouldn't touch this.
	 */
	public void setChangeType (String changeType) {
		this.changeType = changeType;
	}
	
	/**
	 * More or less: {@link org.evolizer.changedistiller.model.classifiers.ChangeType}
	 * Always corresponds to {@link ChangeOperation} values. 
	 * @see {@link ChangeOperation.valueOf(String)}
	 * @see {@link DiffCodes}
	 */
	@Type(type="text")
	public String getChangeType () {
		return this.changeType;
	}
	
	/**
	 * Returns the signature of the method enclosing the left side of the diff, or null
	 * if this is not applicable (e.g., for method inserts) 
	 */
	@Type(type="text")
	public String getLeftMethodSignature () {
		return this.leftMethodSignature;
	}
	
	/**
	 * Returns the signature of the method enclosing the right side of the diff, or null
	 * if this is not applicable (e.g., for field renames)
	 */
	@Type(type="text")
	public String getRightMethodSignature () {
		return this.rightMethodSignature;
	}
	
	/**
	 * Returns the signature of the class enclosing the left side of the diff, or null
	 * if this is not applicable (e.g., in cases of class_insert)
	 */
	@Type(type="text")
	public String getLeftClassSignature () {
		return this.leftClassSignature;
	}
	
	/**
	 * Returns the signature of the class enclosing the right side of the diff, or null
	 * if this is not applicable (e.g., for class insertions) 
	 */
	@Type(type="text")
	public String getRightClassSignature () {
		return this.rightClassSignature;
	}
	
	/**
	 * Returns the signature of the field enclosing the left side of the diff, or null
	 * if this is not applicable (e.g., for diffs inside methods) 
	 */
	@Type(type="text")
	public String getLeftFieldSignature () {
		return this.leftFieldSignature;
	}
	
	/**
	 * Returns the signature of the field enclosing the right side of the diff, or null
	 * if this is not applicable (e.g., for class renames) 
	 */
	@Type(type="text")
	public String getRightFieldSignature () {
		return this.rightFieldSignature;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setRightMethodSignature (String rightMethodSignature) {
		this.rightMethodSignature = rightMethodSignature;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setLeftClassSignature (String leftClassSignature) {
		this.leftClassSignature = leftClassSignature;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setRightClassSignature (String rightClassSignature) {
		this.rightClassSignature = rightClassSignature;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setLeftMethodSignature (String leftMethodSignature) {
		this.leftMethodSignature = leftMethodSignature;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setLeftFieldSignature (String leftFieldSignature) {
		this.leftFieldSignature = leftFieldSignature;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setRightFieldSignature (String rightFieldSignature) {
		this.rightFieldSignature = rightFieldSignature;
	}
	
	/**
	 * Clients shouldn't touch this. This is here for no reason.
	 */
	@Transient
	public String[] getFullChangeType () {
		return this.changeType.split(TAG);
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	@Override
	public void setNodeTo (ASTNode node) {
		this.nodeTo = node;
		if (node != null) {
			this.handleTo = node.toString().trim();
			this.posTo = node.getStartPosition();
			this.lenTo = node.getLength();
		} else {
			this.handleTo = null;
			this.posTo = 0;
			this.lenTo = -1; 
		}
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	@Override
	public void setNodeFrom (ASTNode node) {
		this.nodeFrom = node;
		if (this.nodeFrom != null) {
			this.handleFrom = node.toString().trim();
			this.posFrom = node.getStartPosition();
			this.lenFrom = node.getLength();
		} else {
			this.handleFrom = null;
			this.posFrom = 0;
			this.lenFrom = -1;
		}
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setPosFrom (int posFrom) {
		this.posFrom = posFrom;
	}
	
	/**
	 * The start position of the left node
	 */
	@Basic
	public int getPosFrom () {
		return posFrom;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setLenFrom (int lenFrom) {
		this.lenFrom = lenFrom;
	}
	
	/**
	 * The length of the left node.
	 */
	@Basic
	public int getLenFrom () {
		return lenFrom;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setPosTo (int posTo) {
		this.posTo = posTo;
	}
	
	/**
	 * The start position of the right node.
	 */
	@Basic
	public int getPosTo () {
		return posTo;
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setLenTo (int lenTo) {
		this.lenTo = lenTo;
	}
	
	/**
	 * The length of the right node
	 */
	@Basic
	public int getLenTo () {
		return lenTo;
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isLeftInRange (int off, int len) {
		return isInRange (posFrom, lenFrom, off, len);
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean leftEncloses (int off, int len) {
		return encloses (posFrom, lenFrom, off, len);
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean rightEncloses (int off, int len) {
		return encloses (posTo, lenTo, off, len);
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isRightInRange (int off, int len) {
		return isInRange (posTo, lenTo, off, len);
	}
	
	@Transient
	private boolean isInRange (int diffOff, int diffLen, int off, int len) {
		return diffOff >= off && (diffOff + diffLen) <= (off + len);
	}
	
	@Transient
	private boolean encloses (int diffOff, int diffLen, int off, int len) {
		return isInRange (off, len, diffOff, diffLen);
	}
	
	/**
	 * Clients shouldn't touch this.
	 */
	public void setLeftParent(ASTNode leftParent) {
		this.leftParent = leftParent;
		if (this.leftParent != null) {
			this.leftMethodSignature = getParentMethodSignature(leftParent);
			this.leftClassSignature = getParentClassSignature(leftParent);
			this.leftFieldSignature = getParentFieldSignature(leftParent);
		}
	}

	/**
	 * Clients shouldn't touch this.
	 */
	public void setRightParent(ASTNode rightParent) {
		this.rightParent = rightParent;
		if (this.rightParent != null) {
			this.rightMethodSignature = getParentMethodSignature(rightParent);
			this.rightClassSignature = getParentClassSignature(rightParent);
			this.rightFieldSignature = getParentFieldSignature(rightParent);
		}
	}

	private String getParentMethodSignature (ASTNode node) {
		ASTNode parent = ASTUtil.loopUntilNodeType(node, ASTNode.METHOD_DECLARATION);
		if (parent == null) {
			return null;
		}
		
		MethodDeclaration declaration = (MethodDeclaration) parent;
		
		return BindingUtil.toLongHand(declaration.resolveBinding());
	}
	
	private String getParentFieldSignature (ASTNode node) {
		ASTNode parent = ASTUtil.loopUntilNodeType(node, ASTNode.FIELD_DECLARATION);
		if (parent == null) {
			return null;
		}
		
		FieldDeclaration declaration = (FieldDeclaration) parent;
		List<?> fragments = declaration.fragments();
		
		IVariableBinding binding = null;
		if (fragments.size() > 0) {
			VariableDeclarationFragment frag = (VariableDeclarationFragment) fragments.get(0);
			binding = frag.resolveBinding();
		}
		else {
			ASTNode fragment = ASTUtil.loopUntilNodeType(node, ASTNode.VARIABLE_DECLARATION_FRAGMENT);
			VariableDeclarationFragment frag = (VariableDeclarationFragment) fragment;
			binding = frag.resolveBinding();
		}
		
		return BindingUtil.toLongHand(binding);
	}
	
	private String getParentClassSignature (ASTNode node) {
		ASTNode parent = ASTUtil.getTypeContainer(node);
		if (parent == null) {
			return null;
		}
		
		AbstractTypeDeclaration declaration = (AbstractTypeDeclaration) parent;
		
		return BindingUtil.toLongHand(declaration.resolveBinding());
	}
	
	@Override
	public String toString () {
		StringBuffer buffer = new StringBuffer();
		buffer.append(changeType);
		if (handleFrom != null) {
			buffer.append(" ");
			buffer.append(keepOnlyFirstLine(handleFrom));
		}
		if (handleTo != null) {
			buffer.append(" ");
			buffer.append(keepOnlyFirstLine(handleTo));
		}
		return buffer.toString();
	}
	
	private String keepOnlyFirstLine (String line) {
		int newline = line.indexOf('\n');
		if (newline != -1) {
			line = line.substring(0, newline);
		}
		return line;
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isInsert() {
		return changeType.endsWith("_INSERT");
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isDelete() {
		return changeType.endsWith("_DELETE");
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isUpate() {
		if ( changeType.endsWith("_UPDATE") && ! isOrderingUpdate()) {
			return true;
		}
		if (isStatementUpdate()) {
			return true;
		}
		if (changeType.equals(ChangeOperation.AFFECTED_ONLY_BY_RENAMES.name())) {
			return true;
		}
		
		return false;
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isStatementUpdate () {
		return ChangeOperation.isStatementUpdate(changeType);
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isStatementInsert () {
		return ChangeOperation.isStatementInsert(changeType);
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isStatementDelete () {
		return ChangeOperation.isStatementDelete(changeType);
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isOrderingUpdate () {
		return changeType.endsWith("_ORDERING_UPDATE");
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isMove() {
		return changeType.endsWith("_MOVE");
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isSourceCommentChange () {
		return changeType.startsWith("COMMENT_");
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public boolean isUnclassified () {
		return changeType.equals(ChangeOperation.UNCLASSIFIED.name());
	}
	
	/**
	 * Convenience method.
	 */
	@Transient
	public DiffCatResult copy () {
		DiffCatResult copy = new DiffCatResult();
		copy.changeType = new String(this.changeType);
		copy.detectorId = this.detectorId;
		copy.handleFrom = this.handleFrom;
		copy.handleTo = this.handleTo;
		copy.leftParent = this.leftParent;
		copy.lenFrom = this.lenFrom;
		copy.lenTo = this.lenTo;
		copy.nodeFrom = this.nodeFrom;
		copy.nodeTo = this.nodeTo;
		copy.pathFrom = this.pathFrom;
		copy.pathTo = this.pathTo;
		copy.posFrom = this.posFrom;
		copy.posTo = this.posTo;
		copy.leftClassSignature = this.leftClassSignature;
		copy.leftMethodSignature = this.leftMethodSignature;
		copy.leftFieldSignature = this.leftFieldSignature;
		copy.rightClassSignature = this.rightClassSignature;
		copy.rightMethodSignature = this.rightMethodSignature;
		copy.rightFieldSignature = this.rightFieldSignature;
		copy.repositoryId = this.repositoryId;
		copy.rightParent = this.rightParent;
		copy.transactionId = this.transactionId;
		
		return copy;
	}
	
	@Override
	public int hashCode () 
	{
		int hashCode = 1;
		int prime = 31;
		hashCode *= prime + getTransactionId();
		hashCode *= prime + getHashcode(getChangeType());
		hashCode *= prime + getHashcode(getPathFrom());
		hashCode *= prime + getHashcode(getPathTo());
		hashCode *= prime + getHashcode(getNodeFrom());
		hashCode *= prime + getHashcode(getNodeTo());
		hashCode *= prime + getHashcode(getLeftParent());
		hashCode *= prime + getHashcode(getRightParent());
		hashCode *= prime + (getNodeFrom() == null ? getHashcode (getHandleFrom()) : 1);
		hashCode *= prime + (getNodeTo() == null ? getHashcode (getHandleTo()) : 1);
		hashCode *= prime + (getNodeFrom() == null ? getPosFrom() : 1);
		hashCode *= prime + (getNodeFrom() == null ? getLenFrom() : 1);
		hashCode *= prime + (getNodeTo() == null ? getPosTo() : 1);
		hashCode *= prime + (getNodeFrom() == null ? getLenTo() : 1);
		hashCode *= prime + (getLeftParent() == null ? getHashcode (getLeftMethodSignature()) : 1);
		hashCode *= prime + (getRightParent() == null ? getHashcode (getRightMethodSignature()) : 1);
		hashCode *= prime + (getLeftParent() == null ? getHashcode(getLeftClassSignature()) : 1);
		hashCode *= prime + (getRightParent() == null ? getHashcode(getRightClassSignature()) : 1);
		hashCode *= prime + (getLeftParent() == null ? getHashcode(getLeftFieldSignature()) : 1);
		hashCode *= prime + (getRightParent() == null ? getHashcode(getRightFieldSignature()) : 1);
		
		return hashCode;
	}
	
	private int getHashcode (Object o) {
		if (o == null) {
			return 1;
		}
		
		return o.hashCode();
	}
	
	@Override
	public boolean equals (Object o) {
		if (this == o){
			return true;
		}
		if (o == null || ! o.getClass().equals(getClass())) {
			return false;
		}
		
		DiffCatResult r = (DiffCatResult) o;
		 
		if (r.getTransactionId() != getTransactionId()) {
			return false;
		}
		if (! r.getChangeType().equals(getChangeType())) {
			return false;
		}
		
		if (! areEquals(r.getPathFrom(), getPathFrom())) {
			return false;
		}
		
		if (! areEquals(r.getPathTo(), getPathTo())) {
			return false;
		}
		
		if (! areEquals (r.getNodeFrom(), getNodeFrom())) {
			return false;
		}
		
		// only do handle comparison if nodes are both null
		if (r.getNodeFrom() == null && getNodeFrom() == null) {
			if (! areEquals (r.getHandleFrom(), getHandleFrom())) {
				return false;
			}
			if (r.getPosFrom() != getPosFrom()) {
				return false;
			}
			if (r.getLenFrom() != getLenFrom()) {
				return false;
			}
		}
		
		if (! areEquals (r.getNodeTo(), getNodeTo())) {
			return false;
		}
		
		// only do handle comparison if nodes are both null
		if (r.getNodeTo() == null && getNodeTo() == null) {
			if (! areEquals (r.getHandleTo(), getHandleTo())) {
				return false;
			}
			if (r.getPosTo() != getPosTo()) {
				return false;
			}
			if (r.getLenTo() != getLenTo()) {
				return false;
			}
		}
		
		if (! areEquals (r.getLeftParent(), getLeftParent())) {
			return false;
		}
		
		// only do handle comparison if nodes are both null
		if (r.getLeftParent() == null && getLeftParent() == null) {
			if (! areEquals (r.getLeftMethodSignature(), getLeftMethodSignature())) {
				return false;
			}
			if (! areEquals (r.getLeftClassSignature(), getLeftClassSignature())) {
				return false;
			}
			if (! areEquals (r.getLeftFieldSignature(), getLeftFieldSignature())) {
				return false;
			}
		}
		
		if (! areEquals (r.getRightParent(), getRightParent())) {
			return false;
		}
		
		// only do handle comparison if nodes are both null
		if (r.getRightParent() == null && getRightParent() == null) {
			if (! areEquals (r.getRightMethodSignature(), getRightMethodSignature())) {
				return false;
			}
			if (! areEquals (r.getRightClassSignature(), getRightClassSignature())) {
				return false;
			}
			if (! areEquals (r.getRightFieldSignature(), getRightFieldSignature()));
		}
		
		return true;
	}
	
	private boolean areEquals (Object o1, Object o2) {
		if (o1 == null) {
			if (o2 != null) {
				return false;
			}
			
			return true;
		}
		
		return o1.equals(o2);
	}
	
	@Deprecated
	public void clearTransients () {
		this.nodeFrom = null;
		this.nodeTo = null;
		this.leftParent = null;
		this.rightParent = null;
	}
}
