/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.model;

import org.eclipse.jdt.core.dom.ASTNode;

import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsClassInsertDeleteFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsDocEditFilter;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

/**
 * To use this, please see {@link CollectionUtil}'s filterAndCopy method
 * 
 * @author dkawry
 *
 */
public class DiffCodes {
	
	public static IsChangeOperationFilter METHOD_SIGNATURE_DIFF = 
		new IsChangeOperationFilter( 
			ChangeOperation.METHOD_ACCESSIBILITY_DELETE.name(),
			ChangeOperation.METHOD_ACCESSIBILITY_INSERT.name(),
			ChangeOperation.METHOD_ACCESSIBILITY_UPDATE.name(),
			ChangeOperation.METHOD_NAME_UPDATE.name(),
			ChangeOperation.METHOD_OVERRIDABILITY_MODIFIER_DELETE.name(),
			ChangeOperation.METHOD_OVERRIDABILITY_MODIFIER_INSERT.name(),
			ChangeOperation.METHOD_OVERRIDABILITY_MODIFIER_UPDATE.name(),
			ChangeOperation.METHOD_SIGNATURE_UPDATE.name(),
			ChangeOperation.PARAMETER_DELETE.name(),
			ChangeOperation.PARAMETER_INSERT.name(),
			ChangeOperation.PARAMETER_NAME_UPDATE.name(),
			ChangeOperation.PARAMETER_ORDERING_UPDATE.name(),
			ChangeOperation.PARAMETER_TYPE_TRIVIAL_UPDATE.name(),
			ChangeOperation.PARAMETER_TYPE_UPDATE.name(),
			ChangeOperation.PARAMETER_TYPE_RENAME_UPDATE.name(),
			ChangeOperation.RETURN_TYPE_DELETE.name(),
			ChangeOperation.RETURN_TYPE_INSERT.name(),
			ChangeOperation.RETURN_TYPE_TRIVIAL_UPDATE.name(),
			ChangeOperation.RETURN_TYPE_UPDATE.name(), 
			ChangeOperation.RETURN_TYPE_RENAME_UPDATE.name(),
			ChangeOperation.THROWN_EXCEPTION_DELETE.name(),
			ChangeOperation.THROWN_EXCEPTION_INSERT.name(),
			ChangeOperation.THROWN_EXCEPTION_TRIVIAL_UPDATE.name(),
			ChangeOperation.THROWN_EXCEPTION_RENAME_UPDATE.name());
	
	public static IsChangeOperationFilter FIELD_LEVEL_DIFF = 
		new IsChangeOperationFilter(
			ChangeOperation.ATTRIBUTE_ACCESSIBILITY_DELETE.name(),
			ChangeOperation.ATTRIBUTE_ACCESSIBILITY_INSERT.name(),
			ChangeOperation.ATTRIBUTE_ACCESSIBILITY_UPDATE.name(),
			ChangeOperation.ATTRIBUTE_MODIFIABILITY_DELETE.name(),
			ChangeOperation.ATTRIBUTE_MODIFIABILITY_INSERT.name(),
			ChangeOperation.ATTRIBUTE_TYPE_TRIVIAL_UPDATE.name(),
			ChangeOperation.ATTRIBUTE_TYPE_UPDATE.name(),
			ChangeOperation.ATTRIBUTE_TYPE_RENAME_UPDATE.name(), 
			ChangeOperation.ATTRIBUTE_INIT_DELETE.name(),
			ChangeOperation.ATTRIBUTE_INIT_INSERT.name(),
			ChangeOperation.ATTRIBUTE_INIT_UPDATE.name(),
			ChangeOperation.ATTRIBUTE_NAME_UPDATE.name(),
			ChangeOperation.ATTRIBUTE_SIGNATURE_UPDATE.name());
	
	public static IsChangeOperationFilter CLASS_SIGNATURE_DIFF = 
		new IsChangeOperationFilter(
			ChangeOperation.CLASS_NAME_UPDATE.name(),
			ChangeOperation.PARENT_CLASS_DELETE.name(),
			ChangeOperation.PARENT_CLASS_INSERT.name(),
			ChangeOperation.PARENT_CLASS_UPDATE.name(),
			ChangeOperation.PARENT_CLASS_TRIVIAL_UPDATE.name(),
			ChangeOperation.PARENT_CLASS_RENAME_UPDATE.name(),
			ChangeOperation.PARENT_INTERFACE_DELETE.name(),
			ChangeOperation.PARENT_INTERFACE_INSERT.name(),
			ChangeOperation.PARENT_INTERFACE_UPDATE.name(),
			ChangeOperation.PARENT_INTERFACE_TRIVIAL_UPDATE.name(),
			ChangeOperation.PARENT_INTERFACE_RENAME_UPDATE.name(),
			ChangeOperation.CLASS_DERIVABILITY_DELETE.name(),
			ChangeOperation.CLASS_DERIVABILITY_INSERT.name(),
			ChangeOperation.TYPE_ACCESSIBILITY_DELETE.name(),
			ChangeOperation.TYPE_ACCESSIBILITY_INSERT.name(),
			ChangeOperation.TYPE_ACCESSIBILITY_UPDATE.name());
	
	public static IsChangeOperationFilter ELEMENT_LEVEL_DIFF =
		new IsChangeOperationFilter(
			ChangeOperation.ATTRIBUTE_DELETE.name(),
			ChangeOperation.ATTRIBUTE_INSERT.name(),
			ChangeOperation.FUNCTIONALITY_DELETE.name(),
			ChangeOperation.FUNCTIONALITY_INSERT.name(),
			ChangeOperation.TRIVIAL_CONSTRUCTOR_DELETE.name(),
			ChangeOperation.TRIVIAL_CONSTRUCTOR_INSERT.name()
		);
	
	public static IFilter<DiffCatResult> CLASS_INS_DEL_DIFF =  new IsClassInsertDeleteFilter();
	
	private static IFilter<DiffCatResult> METHOD_DIFF = 
		new IFilter<DiffCatResult> () {
			public boolean passesFilter(DiffCatResult t) {
				if (t.getChangeType().equals(ChangeOperation.CLASS_INSERT.name()) ||
					t.getChangeType().equals(ChangeOperation.CLASS_DELETE.name())) {
					return false;
				}
				if (t.getLeftParent() != null && t.getRightParent() != null) {
					return t.getLeftParent().getNodeType() == ASTNode.METHOD_DECLARATION &&
							t.getRightParent().getNodeType() == ASTNode.METHOD_DECLARATION;
				}
				else {
					// safety precaution
					return false;
				}
			}};
			
	public static IFilter<DiffCatResult> METHOD_BODY_DIFF =
		new IFilter<DiffCatResult> () {
			public boolean passesFilter(DiffCatResult t) {
				if (! METHOD_DIFF.passesFilter(t)) {
					return false;
				}
				if (METHOD_SIGNATURE_DIFF.passesFilter(t)) {
					return false;
				}
				if (DOCUMENT_DIFF.passesFilter(t)) {
					return false;
				}
				
				return true;
			}
		};
	
	public static IsChangeOperationFilter NON_ESSENTIAL_DIFF =
		new IsChangeOperationFilter(
				ChangeOperation.AFFECTED_BY_VAR_EXTRACTION.name(),
				ChangeOperation.AFFECTED_ONLY_BY_RENAMES.name(),
				ChangeOperation.LOCAL_VARIABLE_EXTRACTION_INSERT.name(),
				ChangeOperation.TRIVIAL_METHOD_INVOCATION_DELETE.name(),
				ChangeOperation.TRIVIAL_METHOD_INVOCATION_INSERT.name(),
				ChangeOperation.TRIVIAL_RETURN_STATEMENT_DELETE.name(),
				ChangeOperation.TRIVIAL_RETURN_STATEMENT_INSERT.name(),
				ChangeOperation.TRIVIAL_THIS_EXPRESSION_DELETE.name(),
				ChangeOperation.TRIVIAL_THIS_EXPRESSION_INSERT.name(),
				ChangeOperation.TRIVIAL_THIS_EXPRESSION_UPDATE.name(),
				ChangeOperation.TRIVIAL_IF_STATEMENT_UPDATE.name(),
				ChangeOperation.TRIVIAL_CONSTRUCTOR_DELETE.name(),
				ChangeOperation.TRIVIAL_CONSTRUCTOR_INSERT.name(),
				ChangeOperation.THROWN_EXCEPTION_TRIVIAL_UPDATE.name(),
				ChangeOperation.ATTRIBUTE_TYPE_TRIVIAL_UPDATE.name(),
				ChangeOperation.ATTRIBUTE_TYPE_RENAME_UPDATE.name(),
				ChangeOperation.VARIABLE_DECLARED_TYPE_TRIVIAL_UPDATE.name(),
				ChangeOperation.VARIABLE_DECLARED_TYPE_RENAME_UPDATE.name(),
				ChangeOperation.VARIABLE_NAME_UPDATE.name(),					 
				ChangeOperation.PARAMETER_TYPE_TRIVIAL_UPDATE.name(),
				ChangeOperation.PARAMETER_TYPE_RENAME_UPDATE.name(),
				ChangeOperation.THROWN_EXCEPTION_TRIVIAL_UPDATE.name(),
				ChangeOperation.THROWN_EXCEPTION_RENAME_UPDATE.name(),
				ChangeOperation.RETURN_TYPE_TRIVIAL_UPDATE.name(),
				ChangeOperation.RETURN_TYPE_RENAME_UPDATE.name(),
				ChangeOperation.PARENT_CLASS_TRIVIAL_UPDATE.name(),
				ChangeOperation.PARENT_CLASS_RENAME_UPDATE.name(),
				ChangeOperation.PARENT_INTERFACE_TRIVIAL_UPDATE.name(),
				ChangeOperation.PARENT_INTERFACE_RENAME_UPDATE.name(),
				ChangeOperation.STRING_ARG_SPLIT.name(),
				ChangeOperation.REDUNDANT_CLASS_CAST_UPDATE.name(),
				ChangeOperation.TRIVIAL_ATTRIBUTE_INIT_DELETE.name(),
				ChangeOperation.TRIVIAL_ATTRIBUTE_INIT_INSERT.name(),
				ChangeOperation.TRIVIAL_VARIABLE_INITIALIZE_DELETE.name(),
				ChangeOperation.TRIVIAL_VARIABLE_INITIALIZE_INSERT.name()
		);
		
	public static IFilter<DiffCatResult> ESSENTIAL_DIFF =
		new IFilter<DiffCatResult> () {
			public boolean passesFilter(DiffCatResult t) {
				if (NON_ESSENTIAL_DIFF.passesFilter(t)) {
					return false;
				}
				
				return true;
			}
		};
		
	public static IFilter<DiffCatResult> DOCUMENT_DIFF = new IsDocEditFilter();
	
	public static IFilter<DiffCatResult> METHOD_SIG_REFACTOR_DIFF = new IsChangeOperationFilter(
		ChangeOperation.METHOD_NAME_UPDATE.name(),
		ChangeOperation.METHOD_SIGNATURE_UPDATE.name(),
		ChangeOperation.PARAMETER_DELETE.name(),
		ChangeOperation.PARAMETER_INSERT.name(),
		ChangeOperation.PARAMETER_ORDERING_UPDATE.name(),
		ChangeOperation.PARAMETER_TYPE_UPDATE.name(),
		ChangeOperation.PARAMETER_TYPE_RENAME_UPDATE.name()
	);
}
