/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTFlattener;
import ca.mcgill.cs.swevo.diffcat.util.ast.VariableUsageCollector;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsVarExtractionFilter;
import ca.mcgill.cs.swevo.util.SetUtil;

/**
 * Abstract base class for reconciling diff classifications. 
 * @author dkawry
 */
public abstract class AbstractReconciler {

	protected ASTFlattener flattener;
	
	public AbstractReconciler () {
		flattener = new ASTFlattener();
	}

	public abstract Set<DiffCatResult> reconcile (Set<DiffCatResult> results);
	
	protected Set<DiffCatResult> singleton (DiffCatResult result) {
		return Collections.singleton(result);
	}

	protected List<ASTNode> flattenToNode(DiffCatResult r) {
		ASTNode to   = r.getNodeTo();
		if (to == null) {
			return Collections.emptyList();
		}
		to.accept(flattener);
		List<ASTNode> flatTo = flattener.getFlattenedAndClear();
		return flatTo;
	}

	protected List<ASTNode> flattenFromNode(DiffCatResult r) {
		ASTNode from = r.getNodeFrom();
		if (from == null) {
			return Collections.emptyList();
		}
		from.accept(flattener);
		List<ASTNode> flatFrom = flattener.getFlattenedAndClear();
		return flatFrom;
	}
	
	protected List<Integer> indicesWithNodeType (int nodeType, List<ASTNode> nodes) {
		List<Integer> indices = new ArrayList<Integer>();
		for (int i = 0; i < nodes.size(); i++) {
			if (nodes.get(i).getNodeType() == nodeType) {
				indices.add(i);
			}
		}
		return indices;
	}

	/**
	 * @pre inlineable instanceof VariableDeclarationStatement
	 */
	protected VariableDeclarationFragment getFragment(ASTNode inlineable) {
		return (VariableDeclarationFragment) 
				((VariableDeclarationStatement) inlineable).fragments().get(0);
	}

	protected String constructNodeToGivenExtracts (DiffCatResult affected, Set<DiffCatResult> affectingDiffs)
	{
		String result = affected.getNodeTo().toString();
		
		for (DiffCatResult affecting : affectingDiffs) {
			
			VariableDeclarationFragment frag = getFragment (affecting.getNodeTo());
			SimpleName name = frag.getName();
			Expression expr = frag.getInitializer();
			
			List<Integer> locations = getLocationsOf(result, name.getIdentifier());
			
			if (locations.size() != 1) {
				return affected.getNodeTo().toString();
			}
			
			StringBuffer buf = new StringBuffer();
			buf.append(result.substring(0, locations.get(0)));
			buf.append(expr.toString());
			buf.append(result.substring(locations.get(0) + name.getIdentifier().length()));
			
			result = buf.toString();
		}
		
		return result;
	}
	
	protected String constructNodeFromGivenInlines (DiffCatResult affected, Set<DiffCatResult> affectingDiffs)
	{
		String result = affected.getNodeFrom().toString();
		
		for (DiffCatResult affecting : affectingDiffs) {
			
			VariableDeclarationFragment frag = getFragment (affecting.getNodeFrom());
			SimpleName name = frag.getName();
			Expression expr = frag.getInitializer();
			
			List<Integer> locations = getLocationsOf(result, name.getIdentifier());
			
			if (locations.size() != 1) {
				return affected.getNodeFrom().toString();
			}
			
			StringBuffer buf = new StringBuffer();
			buf.append(result.substring(0, locations.get(0)));
			buf.append(expr.toString());
			buf.append(result.substring(locations.get(0) + name.getIdentifier().length()));
			
			result = buf.toString();
		}
		
		return result;
	}
	
	protected String substituteVariable (MethodDeclaration declaration, SimpleName var, Expression expr)
	{
		String method = declaration.toString();
		
		// find all string locs
		List<Integer> varLocs = getLocationsOf(method, var.getIdentifier());
		
		if (varLocs.isEmpty()) {
			return declaration.toString();
		}
		
		// find locations of initializing expression
		List<Integer> exprLocs = getLocationsOf(method, expr.toString());
		
		if (exprLocs.isEmpty()) {
			return declaration.toString();
		}
		exprLocs = exprLocs.subList(0, 1);
		
		// find closest string loc to where we want to insert
		int bestStringLoc = -1; 
		int bestDistance = Integer.MAX_VALUE;	
		
		for (int varLoc : varLocs) {
			int distance = varLoc - exprLocs.get(0);
			if (distance >= 0 && distance < bestDistance) {
				bestDistance = distance;
				bestStringLoc = varLoc;
			}
		}
		
		StringBuilder bul = new StringBuilder();
		
		// get pre-chunk
		bul.append(method.substring(0, bestStringLoc));
		bul.append(expr.toString());
		bul.append(method.substring(bestStringLoc + var.getLength()));
		
		return bul.toString();
	}
	
	/**
	 * variableName is a variable name and we look for it inside string
	 */
	protected List<Integer> getLocationsOf(String string, String variableName) {
		List<Integer> locs = new ArrayList<Integer>();
		int i = 0;
		while (i < string.length()) {
			int j = string.indexOf(variableName, i);
			if (j == -1) {
				break;
			}
			if (j > 0 && j + variableName.length() < string.length()) {
				char c = string.charAt(j - 1);
				if (Character.isJavaIdentifierPart(c)) {
					i = j + 1;
					continue;
				}
				c = string.charAt(j + variableName.length());
				if (Character.isJavaIdentifierPart(c)) {
					i = j + 1;
					continue;
				}
				locs.add(j);
				i = j + 1;
			}
			else if (j + variableName.length() < string.length()) {
				char c = string.charAt(j + variableName.length());
				if (Character.isJavaIdentifierPart(c)) {
					i = j + 1;
					continue;
				}
				locs.add(j);
				i = j + 1;
			}
			else {
				char c = string.charAt(j - 1);
				if (Character.isJavaIdentifierPart(c)) {
					i = j + 1;
					continue;
				}
				locs.add(j);
				i = j + 1;
			}
		}
		return locs;
	}
	
	protected String replaceAllVarNames (String string, String varName, String repl) {
		
		int lastIndex = 0;
		while (lastIndex < string.length()) {
			int i = string.indexOf(varName, lastIndex);
			if (i == -1) {
				return string;
			}
			char before = 0, after = 0;
			if (i > 0) {
				before = string.charAt(i - 1);
			}
			if (i + varName.length() < string.length()) {
				after = string.charAt(i + varName.length());
			}
			
			lastIndex = i + varName.length();
			if (before != 0) {
				if (Character.isJavaIdentifierPart(before)) {
					continue;
				}
			}
			
			if (after != 0) {
				if (Character.isJavaIdentifierPart(after)) {
					continue;
				}
			}
			
			StringBuffer buf = new StringBuffer();
			buf.append(string.substring(0, i));
			buf.append(repl);
			buf.append(string.substring(i + varName.length()));
			
			string = buf.toString();
			
			lastIndex = i + repl.length();
		}
		
		return string;
	}

	
	/**
	 * Given an "AFFECTED_BY_VAR_EXTRACT", returns all results in changes that
	 * contain a var extraction affecting the AFFECTED_BY_VAR_EXTRACT 
	 * @param result Represents a result affected by one or more var extracts
	 * @param changes A list of candidate changes 
	 * @return all changes in changes affecting result
	 */
	protected Set<DiffCatResult> getAssociatedVarExtracts (DiffCatResult result, Set<DiffCatResult> changes)
	{	
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		changes = SetUtil.filterAndCopy(changes, new IsVarExtractionFilter());
		
		for (DiffCatResult change : changes) {
			
			VariableDeclarationFragment frag = getFragment(change.getNodeTo());
			SimpleName name = frag.getName();
			
			List<SimpleName> occurrences = getOccurrencesOf (result.getNodeTo(), name);
			
			if (occurrences.size() == 1) {
				results.add(change);
			}
		}
		
		return results;
	}
	
	protected List<SimpleName> getOccurrencesOf (ASTNode node, SimpleName name) {
		VariableUsageCollector col = new VariableUsageCollector(name);
		node.accept(col);
		List<SimpleName> list = col.getOccurrences();
		col.clear();
		return list;
	}
}
