/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.refiner.AbstractRefiner;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.util.filter.IFilter;

public class ConstructorInsDelReconciler extends AbstractRefiner {

	private final IFilter<DiffCatResult> isFuncInsDelFilter 
		= new IsChangeOperationFilter(ChangeOperation.FUNCTIONALITY_DELETE.name(), 
									  ChangeOperation.FUNCTIONALITY_INSERT.name());
	
	@Override
	public Set<DiffCatResult> refine(DiffCatResult codeChange) {
		
		if (! isFuncInsDelFilter.passesFilter(codeChange)) {
			return singleton(codeChange);
		}
		
		MethodDeclaration decl = null;
		if (codeChange.getChangeType().equals(ChangeOperation.FUNCTIONALITY_DELETE.name())) {
			decl = (MethodDeclaration) codeChange.getNodeFrom();
		} else {
			decl = (MethodDeclaration) codeChange.getNodeTo();
		}
		
		if (! isEligible(decl)) {
			return singleton(codeChange);
		}
		
		if (codeChange.getChangeType().equals(ChangeOperation.FUNCTIONALITY_DELETE.name())) {
			codeChange.setChangeType(ChangeOperation.TRIVIAL_CONSTRUCTOR_DELETE.name());
		} else {
			codeChange.setChangeType(ChangeOperation.TRIVIAL_CONSTRUCTOR_INSERT.name());
		}
		
		return singleton(codeChange);
	}

	private boolean isEligible (MethodDeclaration decl) {
		// 1. only check public, argument-less constructors
		IMethodBinding binding = decl.resolveBinding();
		// a. constructors
		if (! binding.isConstructor()) {
			return false;
		}
		// b. no arguments
		if (binding.getParameterTypes().length > 0) {
			return false;
		}
		// c. public
		if (! Modifier.isPublic(binding.getModifiers())) {
			return false;
		}
		// 2. only check empty or trivially non-empty constructors
		List<?> statements = decl.getBody().statements();
		for (int i = 0; i < statements.size(); i++) {
			Statement statement = (Statement) statements.get(i);
			// we allow trivial return statements
			if (statement.getNodeType() == ASTNode.RETURN_STATEMENT) {
				ReturnStatement returnStatement = (ReturnStatement) statement;
				if (returnStatement.getExpression() != null) {
					return false;
				}
			}
			// we allow trivial super expressions
			else if (statement.getNodeType() == ASTNode.SUPER_CONSTRUCTOR_INVOCATION) {
				SuperConstructorInvocation sci = (SuperConstructorInvocation) statement;
				if(! sci.arguments().isEmpty()) {
					return false;
				}
			}
			// 
			else {
				// TODO we allow any other random temp activity that has no side effects
				return false;
			}
		}
		// 3. only flag trivial constructors if they are alone
		AbstractTypeDeclaration typeContainer = (AbstractTypeDeclaration) ASTUtil.getTypeContainer(decl);
		List<MethodDeclaration> constructors = ASTUtil.getConstructors(typeContainer);
		
		if (constructors.size() > 1) {
			return false;
		}
		
		// 4. all criteria were met, this deletion is trivial
		return true;
	}
}
