/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.evolizer.changedistiller.treedifferencing.matching.measure.LevenshteinCalculator;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

public abstract class InsDelRenameReconciler extends AbstractReconciler {

	private static final int MAX_LENGTH = 1000;
	private final IFilter<DiffCatResult> delFilter = initDelFilter();
	private final IFilter<DiffCatResult> insFilter = initInsFilter();

	private LevenshteinCalculator calculator = new LevenshteinCalculator();;
	
	@Override
	public Set<DiffCatResult> reconcile(Set<DiffCatResult> codeChanges) {
		
		Set<DiffCatResult> dels = SetUtil.filterAndCopy(codeChanges, delFilter);
		Set<DiffCatResult> inss = SetUtil.filterAndCopy(codeChanges, insFilter);
		
		for (DiffCatResult del : dels) {
			
			Set<DiffCatResult> inRenameRoles = new HashSet<DiffCatResult>();
			
			for (DiffCatResult ins : inss) {
				
				if (! canCompare(del, ins)) {
					continue;
				}
				
				int numTested = 0;
				boolean inRenameRole = true;
				for (DiffCatResult codeChange : codeChanges)
				{
					if (! inRenameRole) {
						continue;
					}
					if (canInvestigate(del, ins, codeChange)) {
						IN_RENAME_ROLE t = inRenameRole(codeChange, del, ins);
						inRenameRole &= t == IN_RENAME_ROLE.YES || t == IN_RENAME_ROLE.NO_ANSWER;
						numTested += t == IN_RENAME_ROLE.NO || t == IN_RENAME_ROLE.YES ? 1 : 0;
					}
				}
				
				if (numTested == 0 || ! inRenameRole) {
					continue;
				}
				
				inRenameRoles.add(ins);
			}
			
			if (inRenameRoles.size() != 1) {
				continue;
			}
			
			// found a rename based on diffs ... wow
			
			DiffCatResult ins = CollectionUtil.getFirst(inRenameRoles);
			
			DiffCatResult rename = new DiffCatResult();
			rename.setChangeType(getRenameOp().name());
			rename.setNodeFrom( extractFrom(del) );
			rename.setLeftParent( extractFrom(del).getParent() );
			rename.setPathFrom(del.getPathFrom());
			rename.setNodeTo( extractTo(ins));
			rename.setRightParent( extractTo(ins).getParent() );
			rename.setPathTo(ins.getPathTo());
			
			codeChanges.remove(del);
			codeChanges.remove(ins);
			inss.remove(ins);
			
			codeChanges.add(rename);
		}
		
		return codeChanges;
	}

	private enum IN_RENAME_ROLE {
		NO_ANSWER, NO, YES;
	}
	
	private IN_RENAME_ROLE inRenameRole (DiffCatResult update, DiffCatResult del, DiffCatResult ins) 
	{
		SimpleName delName = extractFrom(del);
		SimpleName insName = extractTo(ins);
		
		List<SimpleName> delOcs = getOccurrencesOf(update.getNodeFrom(), delName);
		List<SimpleName> insOcs = getOccurrencesOf(update.getNodeTo(), insName);
		
		if (delOcs.size() != insOcs.size()) {
			return IN_RENAME_ROLE.NO;
		}
		
		if (delOcs.size() == 0) {
			return IN_RENAME_ROLE.NO_ANSWER;
		}
		
		String beforeString = update.getNodeFrom().toString();
		String afterString = update.getNodeTo().toString();
		
		double before, after;
		if (beforeString.length() < MAX_LENGTH && afterString.length() < MAX_LENGTH) {
			before = calculator.calculateSimilarity(beforeString, afterString);
		}
		else {
			before = beforeString.equals(afterString) ? 1.0 : 0.1;
		}
		
		String newAfterString = replaceAllVarNames(afterString, insName.getIdentifier(), delName.getIdentifier());
		
		if (beforeString.length() < MAX_LENGTH && newAfterString.length() < MAX_LENGTH) {
			after = calculator.calculateSimilarity(beforeString, newAfterString);
		}
		else {
			after = beforeString.equals(newAfterString) ? 1.0 : 0.0;
		}
		
		if (after < before) {
			return IN_RENAME_ROLE.NO;
		}
		
		return IN_RENAME_ROLE.YES;
	}

	protected abstract IFilter<DiffCatResult> initDelFilter ();
	protected abstract IFilter<DiffCatResult> initInsFilter ();
	
	protected abstract boolean canInvestigate (DiffCatResult del, DiffCatResult ins, DiffCatResult change);
	
	protected abstract boolean canCompare (DiffCatResult del, DiffCatResult ins);
	
	protected abstract SimpleName extractFrom(DiffCatResult del);
	protected abstract SimpleName extractTo(DiffCatResult ins);
	
	protected abstract ChangeOperation getRenameOp ();

}
