/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.Collection;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import ca.mcgill.cs.swevo.diffcat.cd.ChangeDistillerProxy;
import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.model.DiffCodes;
import ca.mcgill.cs.swevo.diffcat.refiner.IDiffRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.VariableDeclarationRefiner;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.diffcat.util.ast.VariableUsageCollector;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.diffcat.util.op.SetPathsOp;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

public class LocalVarExtractionReconciler extends AbstractReconciler {

	protected DiffCatParentMap parentMap;

	private final IsInlineableFilter filter = new IsInlineableFilter();
	
	@Override
	public Set<DiffCatResult> reconcile(Set<DiffCatResult> codeChanges) {
		
		// required to track multiple var extracts per method
		HashSet<DiffCatResult> globalLearned = new HashSet<DiffCatResult>();
		
		HashSet<DiffCatResult> globalUnlearned = new HashSet<DiffCatResult>();
		
		/*
		 * For each method pair
		 * 1. Get all inlineable variable declaration inserts
		 * 2. Inline them
		 * 3. Redo CD diffs for the method
		 *   i. Inlining vars changes one statement
		 *   ii. CD no longer reports a del/ins of that statement
		 *   iii. CD reports it as an update 
		 *   (ii||iii)=> it was a local var extraction
		 *   We report the insert as a var extraction
		 *   We flag the del/ins or update as var extract affected   
		 * 4. Reconcile the new info
		 * 5. Undo inlines
		 */
		
		parentMap = new DiffCatParentMap(codeChanges);
		
		for (LRTuple<ASTNode> methodPair : parentMap.getMethodKeys())
		{
			Collection<DiffCatResult> methodDiffs = parentMap.get(methodPair);
			methodDiffs = CollectionUtil.filterAndCopy(methodDiffs, DiffCodes.METHOD_BODY_DIFF);
			
			Collection<DiffCatResult> inlineVars = CollectionUtil.filterAndCopy(methodDiffs, filter);
			
			for (DiffCatResult inlineVar : inlineVars) {
				
				VariableDeclarationFragment frag = getFragment(inlineVar.getNodeTo());
				SimpleName name = frag.getName();
				if (frag.getInitializer() == null) {
					continue;
				}
				String expr = frag.getInitializer().toString();
				
				VariableUsageCollector col = new VariableUsageCollector(name);
				
				// check if there is an update with the inlineVar
				DiffCatResult update = isInUpdate(col, methodDiffs);
				if (update != null) {
					
					DiffCatResult updateCopy = update.copy();
					updateCopy.setChangeType(ChangeOperation.AFFECTED_BY_VAR_EXTRACTION.name());
					
					DiffCatResult inlineVarCopy = inlineVar.copy();
					inlineVarCopy.setChangeType(ChangeOperation.LOCAL_VARIABLE_EXTRACTION_INSERT.name());
					
					// we've learned this
					globalLearned.add(update);
					globalLearned.add(inlineVar);
					
					// we've unlearned this
					globalUnlearned.add(updateCopy);
					globalUnlearned.add(inlineVarCopy);
					
					continue;
				}
				
				// otherwise must reconcile insert+delete statements
				
				// get new diffs after inlining var
				MethodDeclaration left = (MethodDeclaration) methodPair.getLeft();
				MethodDeclaration right = (MethodDeclaration) methodPair.getRight();
				Set<DiffCatResult> results = inlineAndDiff (left, right, inlineVar);
				results = SetUtil.apply(results, new SetPathsOp(inlineVar.getPathFrom(), inlineVar.getPathTo()));
				
				// check if anything missing from the old results
				Set<DiffCatResult> missingResults = new HashSet<DiffCatResult>();
				for (DiffCatResult methodDiff : methodDiffs) {
					if (! results.contains(methodDiff)) {
						missingResults.add(methodDiff);
					}
				}
				
				// what we learned from this var extract (relevant to the extract itself)
				Set<DiffCatResult> learned = new HashSet<DiffCatResult>();
				
				/*
				 * anything missing was either
				 * i. directly affected by extraction 
				 * ii. indirectly affected, i.e., by 
				 *     the improved textual similarity between code blocks
				 */
				for (DiffCatResult missingResult : missingResults)
				{
					if (missingResult.isDelete())
					{
						ASTNode nodeFrom = missingResult.getNodeFrom();
						
						// directly affected
						if (nodeFrom.toString().contains(expr)) {
							learned.add(missingResult);
						}
					}
					
					if (missingResult.isInsert())
					{
						ASTNode nodeTo = missingResult.getNodeTo();
						if (nodeTo == null) {
							continue;
						}
						nodeTo.accept(col);
						
						// directlyAffected
						if (! col.getOccurrences().isEmpty()) {
							learned.add(missingResult);
						}
						
						
						col.clear();
					}
				}
				
				// check that we learned about the disappearance of an insert/delete pair 
				// referring to the extracted variable
				if (learned.size() != 2) {
					continue;	
				}
				
				// check learned contains exactly one insert and one delete
				int d = 0, i = 0;
				for (DiffCatResult r : learned) {
					if (r.isDelete()) {
						d++;
					} else {
						i++;
					}
				}
				
				if (d != 1 || i != 1) {
					continue;
				}
				
				// we did, so it was a var extraction. 
				// remove everything that we no longer care about
				globalUnlearned.addAll(missingResults);
				
				// learn the relevant new things
				for (DiffCatResult newResult : results) {
					
					// don't learn the broken update statement.. we build this directly later
					if (newResult.isUpate() && newResult.getNodeTo() == null) {
						continue;
					}
					
					// if this change is not new, ignore it
					if (codeChanges.contains(newResult)) {
						continue;
					}
					
					// learn it
					globalLearned.add(newResult);
				}
			
				// construct the var extraction affected
				DiffCatResult first = CollectionUtil.getFirst(learned);
				learned.remove(first);
				DiffCatResult second = CollectionUtil.getFirst(learned);
				learned.remove(second);
				
				// make first del, second ins 
				if (first.getNodeFrom() == null) {
					DiffCatResult temp = first;
					first = second;
					second = temp;
				}
				
				// build it
				DiffCatResult aff = new DiffCatResult();
				aff.setChangeType(ChangeOperation.AFFECTED_BY_VAR_EXTRACTION.name());
				aff.setNodeFrom(first.getNodeFrom());
				aff.setLeftParent(first.getLeftParent());
				aff.setPathFrom(first.getPathFrom());
				aff.setNodeTo(second.getNodeTo());
				aff.setRightParent(second.getRightParent());
				aff.setPathTo(second.getPathTo());
				
				// learn it
				globalLearned.add(aff);
				
				// make a copy of the var extraction
				DiffCatResult inlineVarCopy = inlineVar.copy();
				inlineVarCopy.setChangeType(ChangeOperation.LOCAL_VARIABLE_EXTRACTION_INSERT.name());
				
				// learn it
				globalLearned.add(inlineVarCopy);
				
				// unlearn the original insert
				globalUnlearned.add(inlineVar);
			}
		}
		
		// unlearn everything now
		codeChanges.removeAll(globalUnlearned);
		
		// learn everything now for real
		codeChanges.addAll(globalLearned);
		
		return codeChanges;
	}
	
	private DiffCatResult isInUpdate (VariableUsageCollector col, Collection<DiffCatResult> results)
	{
		DiffCatResult r = null;
		for (DiffCatResult result : results) {
			if (result.isUpate()) {
				result.getNodeTo().accept(col);
				List<SimpleName> ocs = col.getOccurrences();
				col.clear();
				if (ocs.size() == 1) {
					r = result;
				}
			}
		}
		
		return r;
	}
	
	private Set<DiffCatResult> inlineAndDiff (MethodDeclaration left, MethodDeclaration right, DiffCatResult inlineVar) 
	{
		VariableDeclarationFragment frag = getFragment(inlineVar.getNodeTo());
		SimpleName name = frag.getName();
		Expression expr = frag.getInitializer();
		
		String leftString = left.toString();
		String rightString = substituteVariable(right, name, expr);
		
		ChangeDistillerProxy proxy = ChangeDistillerProxy.getProxy();
		
		Set<DiffCatResult> results = proxy.distillMethods (left, leftString, right, rightString);
		
		// necessary to make sure we don't say erroneous var decl shit is missing
		Set<DiffCatResult> refined = refine (results, new VariableDeclarationRefiner());
		
		return refined;
	}
	
	private Set<DiffCatResult> refine (Set<DiffCatResult> results, IDiffRefiner refiner) {
		Set<DiffCatResult> newResults = new HashSet<DiffCatResult>();
		for (DiffCatResult result : results) {
			Set<DiffCatResult> refined = refiner.refine(result);
			newResults.addAll (refined);
		}
		
		return newResults;
	}
	
	/**
	 * To get all inlineable vars easily
	 */
	private class IsInlineableFilter implements IFilter<DiffCatResult>
	{
		private final IFilter<DiffCatResult> filter 
			= new IsChangeOperationFilter(ChangeOperation.VARIABLE_DECLARATION_STATEMENT_INSERT.name());
		
		@Override
		public boolean passesFilter(DiffCatResult t) {
			return isInlinable (t);
		}
		
		private boolean isInlinable (DiffCatResult varInsert) {
			
			if (! filter.passesFilter(varInsert)) {
				return false;
			}
			
			VariableDeclarationStatement varDecl = (VariableDeclarationStatement) varInsert.getNodeTo();
			
			if (varDecl == null) {
				return false;
			}
			
			// we'll just take care of the simple case
			if (varDecl.fragments().size() != 1) {
				return false;
			}
			
			VariableDeclarationFragment frag = (VariableDeclarationFragment) varDecl.fragments().get(0);
			
			SimpleName name = frag.getName();
			VariableUsageCollector collector = new VariableUsageCollector(name);
			varDecl.getParent().accept (collector);
			
			List<SimpleName> occurrences = collector.getOccurrences();
			
			// we'll just handle this one scenario
			if (occurrences.size() != 1) {
				return false;
			}
			
			return true;
		}
	}
	
}
