/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.Collection;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.SimpleName;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

public class ThrownExceptionsReconciler extends AbstractReconciler {

	private IFilter<DiffCatResult> delFilter = new
		IsChangeOperationFilter(ChangeOperation.THROWN_EXCEPTION_DELETE.name());
	
	private IFilter<DiffCatResult> insFilter = new
		IsChangeOperationFilter(ChangeOperation.THROWN_EXCEPTION_INSERT.name());
	
	private IFilter<DiffCatResult> typeUpdate = new
		IsChangeOperationFilter(ChangeOperation.CLASS_NAME_UPDATE.name()); 
	
	@Override
	public Set<DiffCatResult> reconcile(Set<DiffCatResult> codeChanges) {
		
		Set<DiffCatResult> dels = SetUtil.filterAndCopy(codeChanges, delFilter);
		Set<DiffCatResult> inss = SetUtil.filterAndCopy(codeChanges, insFilter);
		Set<DiffCatResult> types = SetUtil.filterAndCopy(codeChanges, typeUpdate);
		
		DiffCatParentMap parentMap = new DiffCatParentMap();
		parentMap.addPairs(dels);
		parentMap.addPairs(inss);
		
		for (LRTuple<ASTNode> methodPair : parentMap.keySet()) {
			
			Collection<DiffCatResult> deletes = CollectionUtil.filterAndCopy(parentMap.get(methodPair), delFilter);
			Collection<DiffCatResult> inserts = CollectionUtil.filterAndCopy(parentMap.get(methodPair), insFilter);
			
			for (DiffCatResult delete : deletes) {
				
				Name delException = (Name) delete.getNodeFrom();
				IBinding delBinding = delException.resolveBinding();
				
				for (DiffCatResult insert : inserts) {
					
					Name insException = (Name) insert.getNodeTo();
					IBinding insBinding = insException.resolveBinding();
					
					// we miss this when fqn/sn combination
					if (BindingUtil.areEqualBindings(delBinding, insBinding)) {
						
						DiffCatResult update = makeUpdate(delete, insert);
						codeChanges.add(update);
						codeChanges.remove(delete);
						codeChanges.remove(insert);
						
						continue;
					}
					
					// we miss this when fqn/sn + rename combination
					if (isAffectedByRenames(delBinding, insBinding, types)) {
						
						DiffCatResult update = makeUpdate(delete, insert);
						update.setChangeType(ChangeOperation.THROWN_EXCEPTION_RENAME_UPDATE.name());
						codeChanges.add(update);
						codeChanges.remove(delete);
						codeChanges.remove(insert);
						
						continue;
					}
				}
				
			}
		}
		
		return codeChanges;
	}

	private DiffCatResult makeUpdate(DiffCatResult delete, DiffCatResult insert) {
		DiffCatResult update = new DiffCatResult();
		update.setChangeType(ChangeOperation.THROWN_EXCEPTION_TRIVIAL_UPDATE.name());
		update.setNodeFrom(delete.getNodeFrom());
		update.setNodeTo(insert.getNodeTo());
		update.setLeftParent (delete.getLeftParent());
		update.setRightParent (insert.getRightParent());
		update.setPathFrom(delete.getPathFrom());
		update.setPathTo(delete.getPathTo());
		return update;
	}

	private boolean isAffectedByRenames (IBinding leftBinding, IBinding rightBinding, Set<DiffCatResult> classUpdates)
	{
		boolean result = false;
		for (DiffCatResult r : classUpdates) {
			SimpleName old = (SimpleName) r.getNodeFrom();
			SimpleName ins = (SimpleName) r.getNodeTo();
			
			IBinding oldBinding = old.resolveBinding();
			IBinding newBinding = ins.resolveBinding();
			
			result |= (oldBinding == leftBinding || newBinding == rightBinding);
		}
		
		return result;
	}
}
