/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.Collection;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Type;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsClassRenameFilter;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

/**
 * Like RenameReconciler, but works exclusively 
 * on attribute, parameter, and local variable type
 * updates. We need to do those first to make sure
 * that entities with *both* a name update *and* a 
 * legitimate type update (i.e., not a rename) produce
 * real updates. Flagging legitimate type updates first
 * makes it easier to reconcile rename-induced diffs in 
 * the rename reconciler  
 * 
 * @author david
 *
 */
public class TypeRenameReconciler extends RenameReconciler {

	private final IsClassRenameFilter classRenames = new IsClassRenameFilter();
	
	private final IFilter<DiffCatResult> typeUpdates = new IFilter<DiffCatResult>() {
		public boolean passesFilter(DiffCatResult t) {
			String changeType = t.getChangeType();
			return 
			changeType.equals(ChangeOperation.ATTRIBUTE_TYPE_UPDATE.name())||
			changeType.equals(ChangeOperation.PARAMETER_TYPE_UPDATE.name())||
			changeType.equals(ChangeOperation.RETURN_TYPE_UPDATE.name())||
			changeType.equals(ChangeOperation.VARIABLE_DECLARED_TYPE_UPDATE.name());
		}
	};
	
	@Override
	public Set<DiffCatResult> reconcile(Set<DiffCatResult> results) {
		
		parentMap = new DiffCatParentMap(results);
		
		Collection<DiffCatResult> changes = parentMap.values();
		Collection<DiffCatResult> renames = CollectionUtil.filterAndCopy(changes, classRenames);
		Collection<DiffCatResult> updates = CollectionUtil.filterAndCopy(changes, typeUpdates);
		
		for (DiffCatResult update : updates) {
			Type nodeFrom = (Type) update.getNodeFrom();
			Type nodeTo = (Type) update.getNodeTo();
			
			ITypeBinding resolveBinding1 = nodeFrom.resolveBinding();
			ITypeBinding resolveBinding2 = nodeTo.resolveBinding();
			
			if (onlyAffectedByRename(resolveBinding1, resolveBinding2, renames)) {
				String changeType = update.getChangeType();
				if (changeType.equals(ChangeOperation.ATTRIBUTE_TYPE_UPDATE.name())) {
					update.setChangeType(ChangeOperation.ATTRIBUTE_TYPE_RENAME_UPDATE.name());
				} else if (changeType.equals(ChangeOperation.PARAMETER_TYPE_UPDATE.name())) {
					update.setChangeType(ChangeOperation.PARAMETER_TYPE_RENAME_UPDATE.name());
				} else if (changeType.equals(ChangeOperation.RETURN_TYPE_UPDATE.name())) {
					update.setChangeType(ChangeOperation.RETURN_TYPE_RENAME_UPDATE.name());
				} else if (changeType.equals(ChangeOperation.VARIABLE_DECLARED_TYPE_UPDATE.name())) {
					update.setChangeType(ChangeOperation.VARIABLE_DECLARED_TYPE_RENAME_UPDATE.name());
				}
			}
		}
		
		return parentMap.values();
	}

}
