/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.refiner;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.BooleanLiteral;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.ParenthesizedExpression;
import org.eclipse.jdt.core.dom.PrefixExpression;
import org.eclipse.jdt.core.dom.QualifiedName;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
import org.eclipse.jdt.core.dom.InfixExpression.Operator;
import org.eclipse.jdt.core.dom.SimpleName;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.reconciler.AbstractReconciler;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.util.CollectionUtil;

public class IfConditionUpdateRefiner extends AbstractRefiner {

	private Stack<Boolean> doAllowMethodInvocs = new Stack<Boolean>();
	private Stack<Boolean> doAllowArrayRefs = new Stack<Boolean>();
	private final AbstractRefiner thisExpr = new ThisExpressionRefiner();
	private final DiffCatResult shell = new DiffCatResult();
	{
		shell.setChangeType(ChangeOperation.IF_STATEMENT_UPDATE.name());
	}
	
	@Override
	public Set<DiffCatResult> refine(DiffCatResult result) {
	
		if (! result.getChangeType().equals(ChangeOperation.IF_STATEMENT_UPDATE.name())) {
			return Collections.singleton(result);
		}
		
		Expression nodeFrom = (Expression) result.getNodeFrom();
		Expression nodeTo = (Expression) result.getNodeTo();
		
		if (areEquivalentExpressions(nodeFrom, nodeTo)) {
			result.setChangeType(ChangeOperation.TRIVIAL_IF_STATEMENT_UPDATE.name());
			return Collections.singleton(result);
		}
		
		return Collections.singleton(result);
	}

	private boolean areEquivalentExpressions (Expression one, Expression two)
	{
		if (one.getNodeType() == two.getNodeType()) {
			if (one.getNodeType() == ASTNode.SIMPLE_NAME || 
				one.getNodeType() == ASTNode.QUALIFIED_NAME || 
				one.getNodeType() == ASTNode.NULL_LITERAL || 
				one.getNodeType() == ASTNode.BOOLEAN_LITERAL || 
				one.getNodeType() == ASTNode.NUMBER_LITERAL || 
				one.getNodeType() == ASTNode.FIELD_ACCESS || 
				one.getNodeType() == ASTNode.SUPER_FIELD_ACCESS || 
				one.getNodeType() == ASTNode.CHARACTER_LITERAL) 
			{
				return one.toString().equals(two.toString());
			}
			
			if (one.getNodeType() == ASTNode.CAST_EXPRESSION || 
				one.getNodeType() == ASTNode.INSTANCEOF_EXPRESSION) {
				return areEquivalentModuloThis(one, two);
			}
			
			if (doAllowMethodInvocs.isEmpty() || doAllowMethodInvocs.peek()) {
				if (one.getNodeType() == ASTNode.SUPER_METHOD_INVOCATION || 
					one.getNodeType() == ASTNode.METHOD_INVOCATION) {
					return areEquivalentModuloThis (one, two);
				}
			}
			
			if (doAllowArrayRefs.isEmpty() || doAllowArrayRefs.peek()) {
				if (one.getNodeType() == ASTNode.ARRAY_ACCESS) {
					return one.toString().equals(two.toString());
				}
			}
			
			if (one.getNodeType() == ASTNode.INFIX_EXPRESSION) {
				return areEquivalent((InfixExpression) one, (InfixExpression) two);
			}
			
			if (one.getNodeType() == ASTNode.PARENTHESIZED_EXPRESSION) {
				ParenthesizedExpression oneP = (ParenthesizedExpression) one;
				ParenthesizedExpression twoP = (ParenthesizedExpression) two;
				
				return areEquivalentExpressions (oneP.getExpression(), twoP.getExpression());
			}
			
			if (one.getNodeType() == ASTNode.PREFIX_EXPRESSION) {
				return areEquivalent( (PrefixExpression) one, (PrefixExpression) two);
			}
		}
		else {
			if (one.getNodeType() == ASTNode.INFIX_EXPRESSION && two.getNodeType() == ASTNode.PREFIX_EXPRESSION) {
				return areEquivalent((InfixExpression) one, (PrefixExpression) two);
			}
			else if (one.getNodeType() == ASTNode.PREFIX_EXPRESSION && two.getNodeType() == ASTNode.INFIX_EXPRESSION) {
				return areEquivalent((InfixExpression) two, (PrefixExpression) one);
			}
			else if (one.getNodeType() == ASTNode.PARENTHESIZED_EXPRESSION) {
				return areEquivalentExpressions ( ((ParenthesizedExpression) one).getExpression(), two);
			}
			else if (two.getNodeType() == ASTNode.PARENTHESIZED_EXPRESSION) {
				return areEquivalentExpressions ( ((ParenthesizedExpression) two).getExpression(), one);
			}
			else if (one.getNodeType() == ASTNode.INFIX_EXPRESSION) {
				return areEquivalent((InfixExpression) one, two);
			}
			else if (two.getNodeType() == ASTNode.INFIX_EXPRESSION) {
				return areEquivalent((InfixExpression) two, one);
			}
			else if (one.getNodeType() == ASTNode.FIELD_ACCESS) {
				if (two.getNodeType() == ASTNode.SIMPLE_NAME) {
					return areEquivalent ((FieldAccess) one, (SimpleName) two);
				}
				else if (two.getNodeType() == ASTNode.QUALIFIED_NAME) {
					return areEquivalent ((FieldAccess) one, (QualifiedName) two);
				}
			}
			else if (two.getNodeType() == ASTNode.FIELD_ACCESS) {
				if (one.getNodeType() == ASTNode.SIMPLE_NAME) {
					return areEquivalent ((FieldAccess) two, (SimpleName) one);
				}
				else if (one.getNodeType() == ASTNode.QUALIFIED_NAME) {
					return areEquivalent ((FieldAccess) two, (QualifiedName) one);
				}
			}
		}
		
		// TODO
		return false;
	}
	
	private boolean areEquivalent (InfixExpression one, InfixExpression two) 
	{
		if (one.getOperator() == two.getOperator()) {
			return areEquivalentSameOperator(one, two);
		}
		else {
			return areEquivalentDifferentOperator(one, two);
		}
	}
	
	private boolean areEquivalent (PrefixExpression one, PrefixExpression two) 
	{
		if (one.getOperator() == two.getOperator()) {
			return areEquivalentExpressions(one.getOperand(), two.getOperand());
		}
		// TODO
		return false;
	}
	
	private boolean areEquivalent (FieldAccess access, SimpleName name) {
		Expression expression = access.getExpression();
		if (expression != null && expression.toString().equals("this")) {
			SimpleName accessName = access.getName();
			if (! accessName.toString().equals(name.toString())) {
				return false;
			}
			
			IVariableBinding a = (IVariableBinding) accessName.resolveBinding();
			IVariableBinding b = (IVariableBinding) name.resolveBinding();
			
			return BindingUtil.areEqual(a.getDeclaringMethod(), b.getDeclaringMethod());
		}
		
		return false;
	}
	
	private boolean areEquivalent (FieldAccess access, QualifiedName name) {
		Expression expression = access.getExpression();
		if (expression != null && expression.toString().equals("this")) {
			SimpleName accessName = access.getName();
			if (! accessName.toString().equals(name.toString())) {
				return false;
			}
			
			IVariableBinding a = (IVariableBinding) accessName.resolveBinding();
			IVariableBinding b = (IVariableBinding) name.resolveBinding();
			
			return BindingUtil.areEqual(a.getDeclaringMethod(), b.getDeclaringMethod());
		}
		else if (expression != null && expression.getNodeType() == ASTNode.FIELD_ACCESS)
		{
			FieldAccess subAccess = (FieldAccess) expression;
			SimpleName accessName = access.getName();
			
			if (! areEquivalentExpressions(accessName, name.getName())) {
				return false;
			}
			
			if (! areEquivalentExpressions(subAccess, name.getQualifier())) {
				return false;
			}
			
			return true;
		}
		
		return false;
	}
	
	private boolean areEquivalent (InfixExpression one, PrefixExpression two)
	{
		if (one.getOperator() == Operator.EQUALS)
		{
			if (two.getOperator() == org.eclipse.jdt.core.dom.PrefixExpression.Operator.NOT) {
				if (one.getLeftOperand().toString().equals("false")) {
					return areEquivalentExpressions (one.getRightOperand(), two.getOperand());
				}
				if (one.getRightOperand().toString().equals("false")) {
					return areEquivalentExpressions (one.getLeftOperand(), two.getOperand());
				}
			}
		}
		else if (one.getOperator() == Operator.NOT_EQUALS)
		{
			if (two.getOperator() == org.eclipse.jdt.core.dom.PrefixExpression.Operator.NOT) {
				if (one.getLeftOperand().toString().equals("true")) {
					return areEquivalentExpressions (one.getRightOperand(), two.getOperand());
				}
				if (one.getRightOperand().toString().equals("true")) {
					return areEquivalentExpressions (one.getLeftOperand(), two.getOperand());
				}
			}
		}
		return false;
	}
	
	private boolean areEquivalent (InfixExpression one, Expression two)
	{
		if (one.getOperator() == Operator.EQUALS)
		{
			if (one.getLeftOperand().toString().equals("true")) {
				return areEquivalentExpressions (one.getRightOperand(), two);
			}
			if (one.getRightOperand().toString().equals("true")) {
				return areEquivalentExpressions (one.getLeftOperand(), two);
			}
		}
		else if (one.getOperator() == Operator.NOT_EQUALS)
		{
			if (one.getLeftOperand().toString().equals("false")) {
				return areEquivalentExpressions (one.getRightOperand(), two);
			}
			if (one.getRightOperand().toString().equals("false")) {
				return areEquivalentExpressions (one.getLeftOperand(), two);
			}
		}
		
		return false;
	}
	
	private boolean areEquivalentSameOperator (InfixExpression one, InfixExpression two) 
	{
		Expression lo1 = one.getLeftOperand();
		Expression lo2 = two.getLeftOperand();
		
		List extendedOperands1 = one.extendedOperands();
		List extendedOperands2 = two.extendedOperands();
		
		// TODO, can be more intelligent here to find more equivalent cases
		if (extendedOperands1.size() + extendedOperands2.size() > 0) {
			return false;
		}
		
		Expression ro1 = one.getRightOperand();
		Expression ro2 = two.getRightOperand();
		
		Operator operator = one.getOperator();
		if (operator.equals(Operator.EQUALS) || 
			operator.equals(Operator.NOT_EQUALS) || 
			operator.equals(Operator.PLUS) || 
			operator.equals(Operator.TIMES)) 
		{
			if (areEquivalentExpressions(lo1, ro2) && areEquivalentExpressions(ro1, lo2)) {
				return true;
			}
			else if (areEquivalentExpressions(lo1, lo2) && areEquivalentExpressions(ro1, ro2)) {
				return true;
			}
		}
		else if (operator.equals(Operator.CONDITIONAL_OR) || operator.equals(Operator.CONDITIONAL_AND))
		{
			if (areEquivalentExpressions(lo1, lo2) && areEquivalentExpressions(ro1, ro2)) {
				return true;
			}
			doAllowMethodInvocs.push(false);
			doAllowArrayRefs.push(false);
			if (areEquivalentExpressions(lo1, ro2) && areEquivalentExpressions(ro1, lo2)) {
				return true;
			}
			doAllowMethodInvocs.pop();
			doAllowArrayRefs.pop();
		}
		
		// TODO 
		return false;
	}
	
	private boolean areEquivalentModuloThis (ASTNode m1, ASTNode m2)
	{
		// cool
		if (m1.toString().equals(m2.toString())) {
			return true;
		}
		
		shell.setNodeFrom(m1);
		shell.setNodeTo(m2);
		
		Set<DiffCatResult> set = thisExpr.refine(shell);
		DiffCatResult refined = CollectionUtil.getFirst(set);
		
		boolean r = false;
		if (refined.getChangeType().equals(ChangeOperation.TRIVIAL_THIS_EXPRESSION_DELETE.name()) || 
			refined.getChangeType().equals(ChangeOperation.TRIVIAL_THIS_EXPRESSION_INSERT.name()))
		{
			r= true;
		}
		
		shell.setChangeType(ChangeOperation.IF_STATEMENT_UPDATE.name());
		shell.setNodeFrom(null);
		shell.setNodeTo(null);
		
		return r;
	}
	
	private boolean areEquivalentDifferentOperator (InfixExpression one, InfixExpression two)
	{
		// TODO
		return false;
	}
}
