/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.refiner;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Name;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsMethodSignatureRefactoringFilter;

/**
 * Handles ChangeDistiller inferred signature updates. ChangeDistiller
 * does not detect thrown exception updates on any method. We detect
 * those in two phases: in {@link SupplDiff} and here. This refiner
 * detects thrown exception updates on those methods that were refactored.
 * SupplDiff cannot detect these, since it has no knowledge of refactored 
 * method signatures. 
 *  
 * @author david
 */
public class MethodSigRefactoringRefiner extends AbstractRefiner {

	private final IsMethodSignatureRefactoringFilter filter;
	
	{
		filter = new IsMethodSignatureRefactoringFilter();
	}

	@SuppressWarnings("rawtypes")
	@Override
	public Set<DiffCatResult> refine (DiffCatResult result) {
		
		if (! filter.passesFilter(result)) {
			return singleton(result);
		}
		
		// we can always do this based on CD
		MethodDeclaration leftDeclaration = (MethodDeclaration) result.getLeftParent();
		MethodDeclaration rightDeclaration = (MethodDeclaration) result.getRightParent();
		
		List leftExceptions = leftDeclaration.thrownExceptions();
		List rightExceptions = rightDeclaration.thrownExceptions();
		
		List<Name> leftExceptions2 = new ArrayList<Name>(); 
		List<Name> rightExceptions2 = new ArrayList<Name>();
		
		for (Object o : leftExceptions) {
			leftExceptions2.add((Name) o);
		}
		
		for (Object o : rightExceptions) {
			rightExceptions2.add((Name) o);
		}
		
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		results.add(result);
		
		refineExceptions(leftExceptions2, rightExceptions2, result, results);
		
		return results;
	}

	private void refineExceptions (List<Name> leftExceptions, List<Name> rightExceptions, DiffCatResult result, Set<DiffCatResult> results)
	{
		List<ITypeBinding> leftExcTypes = new ArrayList<ITypeBinding>();
		List<ITypeBinding> rightExcTypes = new ArrayList<ITypeBinding>();
		
		for (int i = 0; i < leftExceptions.size(); i++) {
			leftExcTypes.add(leftExceptions.get(i).resolveTypeBinding());
		}
		
		for (int i = 0; i < rightExceptions.size(); i++) {
			rightExcTypes.add(rightExceptions.get(i).resolveTypeBinding());
		}
		
		ITypeBinding[] rightExcTypesArr = rightExcTypes.toArray(new ITypeBinding[rightExcTypes.size()]);
		for (int i = 0; i < leftExcTypes.size(); i++) {
			ITypeBinding exc = leftExcTypes.get(i);
			if (! BindingUtil.inCollection(exc, rightExcTypesArr)) {
				// deleted exception
				results.add(makeDelete (leftExceptions.get(i), result));
			}
		}
		
		ITypeBinding[] leftExcTypesArr = leftExcTypes.toArray(new ITypeBinding[leftExcTypes.size()]);
		for (int i = 0; i < rightExcTypes.size(); i++) {
			ITypeBinding exc = rightExcTypes.get(i);
			if (! BindingUtil.inCollection(exc, leftExcTypesArr)) {
				// inserted exception
				results.add(makeInsert (rightExceptions.get(i), result));
			}
		}
	}
	
	private DiffCatResult makeDelete (Name exception, DiffCatResult result) {
		
		DiffCatResult copy = new DiffCatResult();
		copy.setChangeType(ChangeOperation.THROWN_EXCEPTION_DELETE.name());
		copy.setLeftParent(result.getLeftParent());
		copy.setNodeFrom(exception);
		copy.setPathFrom(result.getPathFrom());
		copy.setPathTo(result.getPathTo());
		copy.setRightParent(result.getRightParent());
		copy.setTransactionId(result.getTransactionId());
		return copy;
	}
	
	private DiffCatResult makeInsert (Name exception, DiffCatResult result) {
		
		DiffCatResult copy = new DiffCatResult();
		copy.setChangeType(ChangeOperation.THROWN_EXCEPTION_INSERT.name());
		copy.setLeftParent(result.getLeftParent());
		copy.setNodeTo(exception);
		copy.setPathFrom(result.getPathFrom());
		copy.setPathTo(result.getPathTo());
		copy.setRightParent(result.getRightParent());
		copy.setTransactionId(result.getTransactionId());
		return copy;
	}
}
