/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.refiner;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.CharacterLiteral;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.InfixExpression.Operator;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.StringLiteral;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;

public class StringSplitArgRefiner extends AbstractRefiner {

	@Override
	public Set<DiffCatResult> refine(DiffCatResult result) {
		String changeType = result.getChangeType();
		if (! changeType.equals(ChangeOperation.METHOD_INVOCATION_UPDATE.name()) && 
			! changeType.equals(ChangeOperation.THROW_STATEMENT_UPDATE.name()) && 
			! changeType.equals(ChangeOperation.ASSIGNMENT_UPDATE.name()) && 
			! changeType.equals(ChangeOperation.ATTRIBUTE_INIT_UPDATE.name()) && 
			! changeType.equals(ChangeOperation.RETURN_STATEMENT_UPDATE.name())) 
		{
			return Collections.singleton(result);
		}
		
		ASTNode nodeFrom = result.getNodeFrom();
		ASTNode nodeTo = result.getNodeTo();
		
		if (nodeFrom == null || nodeTo == null) {
			return Collections.singleton(result);	// TODO, bug
		}
		
		if (nodeFrom.getNodeType() == ASTNode.STRING_LITERAL && nodeTo.getNodeType() == ASTNode.INFIX_EXPRESSION)
		{
			StringLiteral lit1 = (StringLiteral) nodeFrom;
			InfixExpression exp2 = (InfixExpression) nodeTo;
			
			if (areSame(lit1, exp2)) {
				result.setChangeType(ChangeOperation.STRING_ARG_SPLIT.name());
				return Collections.singleton(result);
			}
		}
		
		if (nodeFrom.getNodeType() == ASTNode.INFIX_EXPRESSION && nodeTo.getNodeType() == ASTNode.STRING_LITERAL)
		{
			StringLiteral lit2 = (StringLiteral) nodeTo;
			InfixExpression exp1 = (InfixExpression) nodeFrom;
			
			if (areSame(lit2, exp1)) {
				result.setChangeType(ChangeOperation.STRING_ARG_SPLIT.name());
				return Collections.singleton(result);
			}
		}
		
		if (nodeFrom.getNodeType() != nodeTo.getNodeType()) {
			return Collections.singleton(result);
		}
		
		// TODO check for pointless string arg splits
		// see Ant 4001 for example 
		// TODO fix this (see Xerces 1528)
		if (sameModuloStringArgSplits(nodeFrom, nodeTo)) {
			result.setChangeType(ChangeOperation.STRING_ARG_SPLIT.name());
		}
		
		return Collections.singleton(result);
	}
	
	private boolean sameModuloStringArgSplits (ASTNode one, ASTNode two) 
	{
		if (one.getNodeType() == ASTNode.CLASS_INSTANCE_CREATION) {
			ClassInstanceCreation creation1 = (ClassInstanceCreation) one;
			ClassInstanceCreation creation2 = (ClassInstanceCreation) two;
			
			// can't handle this case
			if(creation1.getAnonymousClassDeclaration() != null || 
					creation2.getAnonymousClassDeclaration() != null) 
			{
				return false;
			}
			
			IMethodBinding binding1 = creation1.resolveConstructorBinding();
			IMethodBinding binding2 = creation2.resolveConstructorBinding();
			
			if (! BindingUtil.areEqual(binding1, binding2)) {
				return false;
			}
			
			List<?> args1 = creation1.arguments();
			List<?> args2 = creation2.arguments();
			
			if (! checkArgs(args1, args2)) {
				return false;
			}
			
			Expression expression1 = creation1.getExpression();
			Expression expression2 = creation2.getExpression(); 
			
			if (expression1 == null && expression2 == null) {
				return true;
			}
			
			if (expression1 == null || expression2 == null) {
				return false;
			}
			
			if (expression1.getNodeType() != expression2.getNodeType()) {
				return false;
			}
			
			return sameModuloStringArgSplits(expression1, expression2); 
		}
		else if (one.getNodeType() == ASTNode.METHOD_INVOCATION) {
			
			MethodInvocation invocation1 = (MethodInvocation) one;
			MethodInvocation invocation2 = (MethodInvocation) two;
			
			IMethodBinding binding1 = invocation1.resolveMethodBinding();
			IMethodBinding binding2 = invocation2.resolveMethodBinding();
			
			if (! BindingUtil.areEqual(binding1, binding2)) {
				return false;
			}
			
			List<?> args1 = invocation1.arguments();
			List<?> args2 = invocation2.arguments();
			
			if (! checkArgs(args1, args2)) {
				return false;
			}
			
			Expression expression1 = invocation1.getExpression();
			Expression expression2 = invocation2.getExpression();
			
			if (expression1 == null && expression2 == null) {
				return true;
			}
			
			if (expression1 == null || expression2 == null) {
				return false;
			}
			
			if (expression1.getNodeType() != expression2.getNodeType()) {
				return false;
			}
			
			return sameModuloStringArgSplits(expression1, expression2); 
		}
		else if (one.getNodeType() == ASTNode.EXPRESSION_STATEMENT)
		{
			ExpressionStatement exprStmt1 = (ExpressionStatement) one;
			ExpressionStatement exprStmt2 = (ExpressionStatement) two;
			
			Expression expr1 = exprStmt1.getExpression();
			Expression expr2 = exprStmt2.getExpression();
			
			if (expr1.getNodeType() != expr2.getNodeType()) {
				return false;
			}
			
			return sameModuloStringArgSplits(expr1, expr2);
		}
		else if (one.getNodeType() == ASTNode.RETURN_STATEMENT)
		{
			ReturnStatement retStmt1 = (ReturnStatement) one;
			ReturnStatement retStmt2 = (ReturnStatement) two;
			
			Expression expr1 = retStmt1.getExpression();
			Expression expr2 = retStmt2.getExpression();
			
			if (expr1 == null || expr2 == null) {
				return false;
			}
			
			if (expr1.getNodeType() != expr2.getNodeType()) {
				return false;
			}
			
			return sameModuloStringArgSplits(expr1, expr2);
		}
		else if (one.getNodeType() == ASTNode.INFIX_EXPRESSION)
		{
			return compareInfixes(one, two);
		}
		
		return false;
	}

	public boolean checkArgs(List<?> args1, List<?> args2) {
		for (int i=0; i<args1.size(); i++) {
			Object arg1 = args1.get(i);
			Object arg2 = args2.get(i);
			
			if (! sameModuloStringArgSplits(arg1, arg2)) {
				return false;
			}
		}
		
		return true;
	}

	private boolean sameModuloStringArgSplits(Object arg1, Object arg2) {
		ASTNode node1 = (ASTNode) arg1;
		ASTNode node2 = (ASTNode) arg2;
		
		if (node1.getNodeType() != node2.getNodeType()) {
			
			if (node1.getNodeType() == ASTNode.CHARACTER_LITERAL && node2.getNodeType() == ASTNode.STRING_LITERAL) {
				CharacterLiteral c = (CharacterLiteral) node1;
				StringLiteral l = (StringLiteral) node2;
				
				return compare(c, l);
			}
			
			if (node1.getNodeType() == ASTNode.STRING_LITERAL && node2.getNodeType() == ASTNode.CHARACTER_LITERAL) {
				CharacterLiteral c = (CharacterLiteral) node2;
				StringLiteral l = (StringLiteral) node1;
				
				return compare(c, l);
			}
			
			if (node1.getNodeType() != ASTNode.STRING_LITERAL && node2.getNodeType() != ASTNode.STRING_LITERAL) {
				return node1.toString().equals(node2.toString());
			}
			
			StringLiteral lit = (StringLiteral) (node1.getNodeType() == ASTNode.STRING_LITERAL ? node1 : node2);
			
			if (node1.getNodeType() == ASTNode.INFIX_EXPRESSION || node2.getNodeType() == ASTNode.INFIX_EXPRESSION)
			{
				InfixExpression exp = (InfixExpression) (node1.getNodeType() == ASTNode.INFIX_EXPRESSION ? node1 : node2);
				return areSame(lit, exp);
			}
			
			// TODO more cases - yes, if those themselves are method invocations
			return false;
		}
		else if (node1.getNodeType() == ASTNode.INFIX_EXPRESSION && 
				 node2.getNodeType() == ASTNode.INFIX_EXPRESSION)
		{
			return compareInfixes(node1, node2);
		}
		else {
			// TODO make more precise - yes, if those themselves are method invocations
			return node1.toString().equals(node2.toString());
		}
	}

	private boolean compareInfixes(ASTNode node1, ASTNode node2) {
		InfixExpression exp1 = (InfixExpression) node1;
		InfixExpression exp2 = (InfixExpression) node2;
		
		String s1 = getStringLiteral(exp1);
		String s2 = getStringLiteral(exp2);
		
		if (s1==null || s2==null) {
			return false;
		}
		else {
			return s1.equals(s2);
		}
	}

	private boolean compare(CharacterLiteral c, StringLiteral l) {
		char ch = c.charValue();
		String s = l.getLiteralValue();
		
		if (s.length() == 1 && s.charAt(0) == ch) {
			return true;
		} else {
			return false;
		}
	}

	private boolean areSame(StringLiteral lit, InfixExpression exp) {
		String recovered = recoverString(exp);
		
		if (recovered == null) {
			return false;
		}
		else {
			return lit.getLiteralValue().equals(recovered);
		}
	}

	private String recoverString (InfixExpression exp) {
		
		if (! exp.getOperator().equals(Operator.PLUS)) {
			return null;
		}
		
		Expression leftOperand = exp.getLeftOperand();
		Expression rightOperand = exp.getRightOperand();
		
		List<?> extended = exp.extendedOperands();
		
		StringBuilder s = new StringBuilder();
		String s1 = getStringLiteral(leftOperand);
		if (s1 == null) {
			return null;
		}
		s.append(s1);
		
		String s2 = getStringLiteral(rightOperand);
		if (s2 == null) {
			return null;
		}
		s.append(s2);
		
		boolean failed = false;
		for (Object e : extended) {
			String s3 = getStringLiteral((Expression) e);
			if (s3 == null) {
				failed = true;
			}
			else {
				s.append(s3);
			}
		}
		
		if (failed) {
			return null;
		}
		
		return s.toString();
	}

	public String getStringLiteral(Expression leftOperand) {
		if (leftOperand.getNodeType() == ASTNode.STRING_LITERAL) {
			StringLiteral leftLit = (StringLiteral) leftOperand;
			return leftLit.getLiteralValue();
		} else if (leftOperand.getNodeType() == ASTNode.INFIX_EXPRESSION) {
			String recoverString = recoverString((InfixExpression) leftOperand);
			if (recoverString == null) {
				return null;
			}
			return recoverString;
		} else if (leftOperand.getNodeType() == ASTNode.CHARACTER_LITERAL) {
			CharacterLiteral leftLit = (CharacterLiteral) leftOperand;
			return leftLit.charValue() + "";
		} else {
			return null;
		}
	}
}
