/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.refiner;

import java.util.Collection;
import java.util.Collections;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.IMethodBinding;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.ast.MethodInvocationCollector;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.util.CollectionUtil;

/**
 * Further classifies method invocations and super constructor invocations
 * into two groups: 
 *  - those equivalent to the (trivial) call to super( )
 *  - those that are not
 *  
 * The distinction is relevant because insertions and deletions of super()
 * are neurotic and not actually useful.     
 *     
 * @author david
 */
public class SuperConstructorRefiner extends AbstractRefiner {

	private final IsChangeOperationFilter filter = 
		new IsChangeOperationFilter(
			ChangeOperation.SUPER_CONSTRUCTOR_INVOCATION_DELETE.name(),
			ChangeOperation.SUPER_CONSTRUCTOR_INVOCATION_INSERT.name(),
			ChangeOperation.SUPER_METHOD_INVOCATION_DELETE.name(),
			ChangeOperation.SUPER_METHOD_INVOCATION_INSERT.name(),
			ChangeOperation.METHOD_INVOCATION_DELETE.name(),
			ChangeOperation.METHOD_INVOCATION_INSERT.name()
		); 

	@Override
	public Set<DiffCatResult> refine(DiffCatResult result) {
		
		if (! filter.passesFilter(result)) {
			return singleton (result);
		}
		
		MethodInvocationCollector collector = new MethodInvocationCollector();
		
		if (result.isDelete()) {
			
			ASTNode node = result.getNodeFrom();
			if (node == null) {
				return singleton(result);
			}
			if (node.toString().trim().equals("super();")) {
				result.setChangeType(ChangeOperation.TRIVIAL_METHOD_INVOCATION_DELETE.name());
			}
			node.accept(collector);
			Collection<IMethodBinding> bindings = collector.getMethodCalls();
			
			if (! isNonTrivial(bindings)) {
				result.setChangeType(ChangeOperation.TRIVIAL_METHOD_INVOCATION_DELETE.name());
			}
		}
		else {
			
			ASTNode node = result.getNodeTo();
			if (node == null) {
				return singleton(result);
			}
			if (node.toString().trim().equals("super();")) {
				result.setChangeType(ChangeOperation.TRIVIAL_METHOD_INVOCATION_INSERT.name());
			}
			node.accept(collector);
			Collection<IMethodBinding> bindings = collector.getMethodCalls();
			
			if (! isNonTrivial(bindings)) {
				result.setChangeType(ChangeOperation.TRIVIAL_METHOD_INVOCATION_INSERT.name());
			}
		}
		
		return Collections.singleton(result);
	}

	private boolean isNonTrivial (Collection<IMethodBinding> bindings) {
		if (bindings.size() < 1) {
			return true;
		}
		if (bindings.size() > 1) {
			return true;
		}
		IMethodBinding binding = CollectionUtil.getFirst(bindings);
		if( binding == null) {
			return true;
		}
		return ! binding.isDefaultConstructor();
	}
}
