/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.refiner;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;

public class VariableDeclarationRefiner extends AbstractRefiner {

	public Set<DiffCatResult> refine(DiffCatResult result) {
		
		Set<DiffCatResult> refined = new HashSet<DiffCatResult>();
		
		if (result.getChangeType().equals(ChangeOperation.VARIABLE_DECLARATION_STATEMENT_UPDATE.name())) {
			VariableDeclarationStatement varLeftStmt = (VariableDeclarationStatement) result.getNodeFrom();
			VariableDeclarationStatement varRightStmt = (VariableDeclarationStatement) result.getNodeTo();
			
			refine (result, varLeftStmt, varRightStmt, refined);
		}
		else if (result.getChangeType().equals(ChangeOperation.CATCH_CLAUSE_UPDATE.name()) ||
				result.getChangeType().equals(ChangeOperation.CATCH_CLAUSE_ORDERING_UPDATE.name())) 
		{
			SingleVariableDeclaration leftClause = (SingleVariableDeclaration) result.getNodeFrom();
			SingleVariableDeclaration rightClause = (SingleVariableDeclaration) result.getNodeTo();
			
			if (leftClause == null || rightClause == null) {
				refined.add(result);
				// TODO, Ant#2622 
			}
			else {
				refine (result, leftClause, rightClause, refined);
			}
		}
		else {
			refined.add(result);
		}
		
		return refined;
	}
	
	private void refine(DiffCatResult result, SingleVariableDeclaration left, SingleVariableDeclaration right, Set<DiffCatResult> refined) {
		
		SimpleName leftName = left.getName();
		SimpleName rightName = right.getName();
		
		if (! leftName.toString().equals(rightName.toString())) {
			
			DiffCatResult copy = result.copy();
			copy.setChangeType(ChangeOperation.VARIABLE_NAME_UPDATE.name());
			copy.setNodeFrom(leftName);
			copy.setNodeTo(rightName);
			refined.add(copy);
		}
		else
		{
			refined.add(result);
		}
		
		return;
	}
	
	@SuppressWarnings("rawtypes")
	private void refine (DiffCatResult r, VariableDeclarationStatement left, VariableDeclarationStatement right, Set<DiffCatResult> refined)
	{
		if (left == null) {
			refined.add(r);
			return;
		}
		List leftFrags = left.fragments();
		
		if (right == null) {
			refined.add(r);
			return;
		}
		List rightFrags = right.fragments();
		
		// too complicated.. for now
		if (leftFrags.size() != 1 || rightFrags.size() != 1) {
			refined.clear();
			refined.add(r);
			return;
		} 
		
		Type leftType = left.getType();
		Type rightType = right.getType();
		
		// we checked that they are the same size
		String varLeftName = "";
		String varRightName = "";
		
		Expression varLeftInit = null;
		Expression varRightInit = null;
		
		VariableDeclarationFragment l = (VariableDeclarationFragment) leftFrags.get(0);
		VariableDeclarationFragment rh = (VariableDeclarationFragment) rightFrags.get(0);
		
		varLeftName = l.getName().toString();
		varRightName = rh.getName().toString();
		
		Expression lInit = l.getInitializer();
		Expression rInit = rh.getInitializer();
		
		varLeftInit = lInit;
		varRightInit = rInit;
			
		Expression lExpr = varLeftInit;
		Expression rExpr = varRightInit;
		
		// declared type update
		if (! leftType.toString().equals(rightType.toString())) {
			
			SimpleName lName = l.getName();
			SimpleName rName = rh.getName();
			
			ITypeBinding lType = lName.resolveTypeBinding();
			ITypeBinding rType = rName.resolveTypeBinding();
			
			if (BindingUtil.areEqual(lType, rType)) {
				DiffCatResult copy = r.copy();
				copy.setChangeType(ChangeOperation.VARIABLE_DECLARED_TYPE_TRIVIAL_UPDATE.name());
				copy.setNodeFrom(leftType);
				copy.setNodeTo(rightType);
				refined.add(copy);	// e.g., ant 2147, 2149
			}
			
			// but check its not a []'ing switch!! (Ant #2526)
			else if (leftType.toString().endsWith("[]") && ! rightType.toString().endsWith("[]") 
				&& ! varLeftName.endsWith("[]") && varRightName.endsWith("[]"))
				{
					DiffCatResult copy = r.copy();
					copy.setChangeType(ChangeOperation.VARIABLE_DECLARED_TYPE_TRIVIAL_UPDATE.name());
					copy.setNodeFrom(leftType);
					copy.setNodeTo(rightType);
					refined.add(copy); 
				}
			
			else if (! leftType.toString().endsWith("[]") && rightType.toString().endsWith("[]")
				&& varLeftName.endsWith("[]") && ! varRightName.endsWith("[]")) 
			{
				DiffCatResult copy = r.copy();
				copy.setChangeType(ChangeOperation.VARIABLE_DECLARED_TYPE_TRIVIAL_UPDATE.name());
				copy.setNodeFrom(leftType);
				copy.setNodeTo(rightType);
				refined.add(copy);
			}
				 
			else {
				DiffCatResult copy = r.copy();
				copy.setChangeType(ChangeOperation.VARIABLE_DECLARED_TYPE_UPDATE.name());
				copy.setNodeFrom(leftType);
				copy.setNodeTo(rightType);
				refined.add(copy);
			}
		}
		
		DiffCatResult copy = null;
			
		if (lExpr == null && rExpr == null) { 
			// nothing to worry about for this part
		}
		else if ( (lExpr == null && rExpr != null) ||
			 (lExpr != null && rExpr == null) ||
			 (! lExpr.toString().equals(rExpr.toString())) ) 
		{
			// found possible expr update 
			copy = r.copy();
			
			if (lExpr == null) {
				copy.setChangeType(ChangeOperation.VARIABLE_INITIALIZE_INSERT.name());
				copy.setNodeFrom(lExpr);
				copy.setNodeTo(rExpr);
				refined.add(copy);
			}
			else if (rExpr == null) {
				copy.setChangeType(ChangeOperation.VARIABLE_INITIALIZE_DELETE.name());
				copy.setNodeFrom(lExpr);
				copy.setNodeTo(rExpr);
				refined.add(copy);
			} else {
				copy.setChangeType(ChangeOperation.VARIABLE_INITIALIZE_UPDATE.name());
				copy.setNodeFrom(lExpr);
				copy.setNodeTo(rExpr);
				refined.add(copy);
			}
		}
			
		String leftName = varLeftName;
		String rightName = varRightName;
		
		if (!leftName.equals(rightName)) {
			// we'll call it variable rename now
			copy = r.copy();
			copy.setChangeType(ChangeOperation.VARIABLE_NAME_UPDATE.name());
			copy.setNodeFrom(l.getName());
			copy.setNodeTo(rh.getName());
			refined.add(copy);
		}
		
		// check insert/delete of final modifier
		int leftModifiers = left.getModifiers();
		int rightModifiers = right.getModifiers();
		
		// final modifier delete
		if (Modifier.isFinal(leftModifiers) && ! Modifier.isFinal(rightModifiers))
		{
			copy = r.copy();
			copy.setChangeType(ChangeOperation.VARIABLE_MODIFIABILITY_INSERT.name());
			copy.setNodeTo(((VariableDeclarationFragment) left.fragments().get(0)).getName());
			refined.add(copy);
		}
		
		// final modifier insert
		if(Modifier.isFinal(rightModifiers) && ! Modifier.isFinal(leftModifiers))
		{
			copy = r.copy();
			copy.setChangeType(ChangeOperation.VARIABLE_MODIFIABILITY_DELETE.name());
			copy.setNodeFrom(((VariableDeclarationFragment) right.fragments().get(0)).getName());
			refined.add(copy);
		}
	}
}
