/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.util.binding;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;

import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IEqualityFilter;
import ca.mcgill.cs.swevo.util.filter.IFilter;

public class BindingUtil {

	public final static String[] UNKNOWN_TYPE_BINDING = new String[] {"UNKNOWNP", "UNKNOWN"};
	
	public static String[] getTypeFQN (ITypeBinding binding) {
		if (binding == null) {
			return UNKNOWN_TYPE_BINDING;
		}
		return binding.getQualifiedName().split(ASTUtil.NON_REGEX_DOT);
	}

	public static String[] getTypeNoGenericsFQN (ITypeBinding binding) {
		if (binding == null) {
			return UNKNOWN_TYPE_BINDING;
		}
		
		String qualName = removeGenerics(binding.getQualifiedName());
		return qualName.split(ASTUtil.NON_REGEX_DOT);
	}
	
	public static String[] getTypeNoGenericsFQNIgnoreBrackets (ITypeBinding binding) {
		if (binding == null) {
			return UNKNOWN_TYPE_BINDING;
		}
		String qualName = removeGenerics(binding.getQualifiedName());
		return removeBrackets(qualName).split(ASTUtil.NON_REGEX_DOT);
	}
	
	public static String removeBrackets (String s) {
		return s.replaceAll("\\[", "").replaceAll("\\]","");
	}
	
	/**
	 * Removes all '<' '>' and things contained btween.
	 * @param args
	 * @return
	 */
	public static String removeGenerics (String args) {
		while (args.contains("<")) {
			int lIndex = args.indexOf('<');
			int numSeen = 1;
			for (int rIndex = lIndex + 1; rIndex < args.length(); rIndex++) {
				char c = args.charAt(rIndex);
				if (c == '<') {
					numSeen++;
				}
				else if (c == '>') {
					numSeen--;
				}
				
				if (numSeen == 0) {
					args = args.substring(0, lIndex) + args.substring(rIndex + 1);
					break;
				}
			}
		}
		
		return args;
	}
	
	public static String translateVarArgs (String args) {
		return args.replaceAll("\\.\\.\\.", "[]");
	}

	public static String[] getDeclaringType (IMethodBinding binding) {
		return getTypeFQN(binding.getDeclaringClass());
	}

	public static String getMethodName (IMethodBinding binding) {
		if (binding == null) {
			return "UNKNOWN";
		}
		return binding.getName();
	}

	public static String[][] getParamTypes (IMethodBinding binding) {
		ITypeBinding[] params = binding.getParameterTypes();
		String[][] result = new String[params.length][];
		for (int i = 0; i < params.length; i++) {
			result[i] = getTypeNoGenericsFQN(params[i]);
		}
		
		return result;
	}

	public static String[][] getParamTypesIgnoreBrackets (IMethodBinding binding) {
		ITypeBinding[] params = binding.getParameterTypes();
		String[][] result = new String[params.length][];
		for (int i = 0; i < params.length; i++) {
			result[i] = getTypeNoGenericsFQNIgnoreBrackets(params[i]);
		}
		
		return result;
	}

	public static boolean validateBinding (ITypeBinding binding) {
		return binding != null && binding.getName() != null;
	}

	public static boolean validateBinding (IVariableBinding binding) {
		
		if (binding == null) {
			return false;
		}
		
		ITypeBinding declaringClass = binding.getDeclaringClass();
		IMethodBinding declaringMethod = binding.getDeclaringMethod();
		
		if (declaringClass == null && declaringMethod == null && ! binding.getName().equals("length")) {
			return false;
		}
		
		if (declaringClass != null) {
			if (! BindingUtil.validateBinding(declaringClass)) {
				return false;
			}
		}
		
		if (declaringMethod != null) {
			if (! BindingUtil.validateBinding(declaringMethod)) {
				return false;
			}
		}		
		
		String name = binding.getName();
		
		if (name == null) {
			return false;
		}
		
		ITypeBinding declaredType = binding.getType();
		
		if (! BindingUtil.validateBinding(declaredType)) {
			return false;
		}
		
		return true;
	}

	public static boolean validateBinding (IMethodBinding binding) {
		
		if (binding == null) {
			return false;
		}
		
		ITypeBinding declaringClass = binding.getDeclaringClass();
		
		if (! BindingUtil.validateBinding(declaringClass)) {
			return false;
		}
		
		String name = binding.getName();
		
		if (name == null) {
			return false;
		}
		
		ITypeBinding[] parameters = binding.getParameterTypes();
		
		if (parameters == null) {
			return false;
		}
		
		for (int i = 0; i < parameters.length; i++) {
			if (! BindingUtil.validateBinding(parameters[i])) {
				return false;
			}
		}
		
		return true;
	}

	public static <T extends IBinding> IEqualityFilter<T> makeEqualityBindingFilter (final T toFind) {
		
		if (toFind == null) {
			return new IEqualityFilter<T>() {
				@Override
				public boolean passesFilter(T t) {
					return false;
				}

				@Override
				public T getComparison() {
					return null;
				}
			};
		}
		
		return new IEqualityFilter<T>() {

			@Override
			public boolean passesFilter(T t) {
				return areEqualBindings(t, toFind);
			}

			@Override
			public T getComparison() {
				return toFind;
			}
		};
	}
	
	public static <T extends IBinding> IFilter<T> makeInequalityBindingFilter (final T toFind) {
		return new IFilter<T>() {

			@Override
			public boolean passesFilter(T t) {
				return ! areEqualBindings(t, toFind);
			}
		};
	}
	
	public static <T extends IBinding> IFilter<T> makeInequalityBindingFilter (final Collection<T> toFind) {
		return new IFilter<T>() {

			@Override
			public boolean passesFilter(T t) {
				return ! inCollection(t, toFind.toArray(new IBinding[toFind.size()]));
			}
		};
	}
	
	public static boolean areEqualBindings (IBinding binding1, IBinding binding2) {
		if (binding1 instanceof IMethodBinding && binding2 instanceof IMethodBinding) {
			return areEqual ( (IMethodBinding) binding1, (IMethodBinding) binding2);
		}
		if (binding1 instanceof IVariableBinding && binding2 instanceof IVariableBinding) {
			return areEqual ( (IVariableBinding) binding1, (IVariableBinding) binding2);
		}
		if (binding1 instanceof ITypeBinding && binding2 instanceof ITypeBinding) {
			return areEqual ( (ITypeBinding) binding1, (ITypeBinding) binding2);
		}
		
		return false;
	}
	
	public static boolean areEqual (IMethodBinding binding1, IMethodBinding binding2) {
		
		if (binding1 == binding2) {
			return true;
		}
		
		if (binding1 == null || binding2 == null) {
			return false;
		}
		
		if (! BindingUtil.areEqual (binding1.getDeclaringClass(), binding2.getDeclaringClass())) {
			return false;
		}
		if (! binding1.getName().equals(binding2.getName())) {
			return false;
		}
		if (! BindingUtil.areEqual (binding1.getParameterTypes(), binding2.getParameterTypes())) {
			return false;
		}
		
		return true;
	}
	
	/**
	 * Utility method that checks whether two method bindings have the same name and parameters.
	 */
	public static boolean areWeaklyOverriding (IMethodBinding binding1, IMethodBinding binding2) {
		
		if (binding1 == null || binding2 == null) {
			return false;
		}
		
		if (! binding1.getName().equals(binding2.getName())) {
			return false;
		}
		
		return BindingUtil.areEqual (binding1.getParameterTypes(), binding2.getParameterTypes());
		
	}

	public static boolean areEqualVariables (Collection<IVariableBinding> bindings1, Collection<IVariableBinding> bindings2) {
		IVariableBinding[] b1 = bindings1.toArray(new IVariableBinding[bindings1.size()]);
		IVariableBinding[] b2 = bindings2.toArray(new IVariableBinding[bindings2.size()]);
		
		return BindingUtil.areEqual (b1, b2);
	}

	public static boolean areEqualMethods (Collection<IMethodBinding> bindings1, Collection<IMethodBinding> bindings2) {
		IMethodBinding[] b1 = bindings1.toArray(new IMethodBinding[bindings1.size()]);
		IMethodBinding[] b2 = bindings2.toArray(new IMethodBinding[bindings2.size()]);
		
		return BindingUtil.areEqual (b1, b2);
	}

	public static boolean hasIntersection (Collection<? extends IBinding> col1, Collection<? extends IBinding> col2) {
		boolean b=false;
		for (IBinding i1 : col1) {
			for (IBinding i2 : col2) {
				b = !b ? areEqualBindings(i1, i2) : b;
			}
		}
		return b;
	}
	
	public static Set<IBinding> intersection (Collection<? extends IBinding> col1, Collection<? extends IBinding> col2) {
		return intersection(
				col1.toArray(new IBinding[col1.size()]), col2.toArray(new IBinding[col2.size()]));
	}
	
	public static Set<IBinding> intersection (IBinding[] col1, IBinding[] col2) {
		if (col1.length == 0 || col2.length == 0) {
			return Collections.emptySet();
		}
		IBinding[] intersection = new IBinding[Math.max(col1.length, col2.length)];
		int cur = 0;
		// we need to check two things: i) already in set, ii) in col2?
		for (int i = 0; i < col1.length; i++) {
			boolean alreadyHave = false;
			for (int j = 0; j < cur; j++) {
				if (areEqualBindings(intersection[j], col1[i])) {
					alreadyHave = true;
					break;
				}
			}
			if (alreadyHave) {
				continue;
			}
			for (int j = 0; j < col2.length; j++) {
				if (areEqualBindings(col1[i], col2[j])) {
					intersection[cur] = col1[i];
					cur++;
				}
			}
		}
		
		Set<IBinding> bindings = new HashSet<IBinding>();
		for (int i = 0; i < cur; i++) {
			bindings.add(intersection[i]);
		}
		
		return bindings;
	}
	
	public static <T extends IBinding> boolean inCollection (T binding, IBinding[] bindings) {
		for (int i = 0; i < bindings.length; i++) {
			if (BindingUtil.areEqualBindings(binding, bindings[i])) {
				return true;
			}
		}
		return false;
	}
	
	public static boolean inCollection (IMethodBinding binding, IMethodBinding[] bindings) {
		for (int i = 0; i < bindings.length; i++) {
			if (BindingUtil.areEqual(binding, bindings[i])) {
				return true;
			}
		}
		return false;
	}

	public static boolean inCollection (IVariableBinding binding, IVariableBinding[] bindings) {
		for (int i = 0; i < bindings.length; i++) {
			if (BindingUtil.areEqual(binding, bindings[i])) {
				return true;
			}
		}
		return false;
	}

	public static boolean inCollection (ITypeBinding binding, ITypeBinding[] bindings) {
		for (int i = 0; i < bindings.length; i++) {
			if (BindingUtil.areEqual(binding, bindings[i])) {
				return true;
			}
		}
		return false;
	}

	public static boolean areEqual (IVariableBinding[] b1, IVariableBinding[] b2) {
		if (b1.length != b2.length) {
			return false;
		}
		
		for (int i = 0; i < b1.length; i++) {
			if (! BindingUtil.areEqual (b1[i], b2[i])) {
				return false;
			}
		}
		
		return true;
	}

	public static boolean areEqual (IMethodBinding[] b1, IMethodBinding[] b2) {
		if (b1.length != b2.length) {
			return false;
		}
		
		for (int i = 0; i < b1.length; i++) {
			if (! areEqual (b1[i], b2[i])) {
				return false;
			}
		}
		
		return true;
	}

	public static boolean areEqual (IVariableBinding binding1, IVariableBinding binding2) {
		
		if (binding1 == binding2) {
			return true;
		}
		
		if (binding1 == null || binding2 == null) {
			return false;
		}
		
		if (! binding1.getName().equals(binding2.getName())) {
			return false;
		}
		if (! areEqual (binding1.getDeclaringClass(), binding2.getDeclaringClass())) {
			return false;
		}
		if (! areEqual (binding1.getDeclaringMethod(), binding2.getDeclaringMethod())) {
			return false;
		}
		
		// XXX this method will say all references to 'length' of an array are the same... yikes 
		
		// don't need to compare on the type of the variable! 
		
		return true;
	}

	public static boolean areEqual (ITypeBinding[] bindings1, ITypeBinding[] bindings2) {
		if (bindings1.length != bindings2.length) {
			return false;
		}
		for (int i = 0; i < bindings1.length; i++) {
			if (! BindingUtil.areEqual (bindings1[i], bindings2[i])) {
				return false;
			}
		}
		return true;
	}

	public static boolean areEqual (ITypeBinding binding1, ITypeBinding binding2) {
		
		if (binding1 == binding2) {
			return true; // OMFG...
		}
		
		if (binding1 == null || binding2 == null) {
			return false;
		}
		
		// TODO is this correct? 
		if (binding1.isTypeVariable() != binding2.isTypeVariable()) {
			return false;
		}
		
		
		if (binding1.isTypeVariable()) {
			ITypeBinding erasure1 = binding1.getErasure();
			ITypeBinding erasure2 = binding2.getErasure();
			
			return areEqual(erasure1, erasure2);
		}
		// ----------------------- 
		
		String qualName1 = binding1.getQualifiedName();
		String qualName2 = binding2.getQualifiedName();
		
		if(! qualName1.equals(qualName2)) {
			return false;
		}
		
		if (qualName1.endsWith("UNKNOWN") || qualName2.endsWith("UNKNOWN")) {
			return false;
		}
		
		if (qualName1.startsWith("UNKNOWNP") || qualName2.startsWith("UNKNOWNP")) {
			return false;
		}
		
		if (! areEqual (binding1.getDeclaringClass(), binding2.getDeclaringClass())) {
			return false;
		}
		
		if (! areEqual (binding1.getDeclaringMethod(), binding2.getDeclaringMethod())) {
			return false;
		}
		
		return true;
	}
	
	public static String toLongHand (IBinding binding) {
		if (binding instanceof ITypeBinding) {
			return toLongHand ((ITypeBinding) binding);
		}
		else if (binding instanceof IVariableBinding) {
			return toLongHand ((IVariableBinding) binding);
		}
		else if (binding instanceof IMethodBinding){
			return toLongHand ((IMethodBinding) binding);
		}
		else {
			return null;
		}
	}
	
	public static String toLongHand (ITypeBinding binding) {
		return binding.getQualifiedName();
	}
	
	public static String toLongHand (IVariableBinding binding) {
		if (binding.getDeclaringClass() != null) {
			return toLongHand (binding.getDeclaringClass()) + "." + binding.getName();
		}
		else if (binding.getDeclaringMethod() != null){
			return toLongHand (binding.getDeclaringMethod()) + "." + binding.getName();
		} 
		else {
			return null;
		}
	}
	
	public static String toLongHand (IMethodBinding binding) {
		StringBuffer buf = new StringBuffer();
		String declClass = toLongHand(binding.getDeclaringClass());
		buf.append(declClass);
		buf.append(".");
		buf.append(binding.getName());
		buf.append("(");
		ITypeBinding[] params = binding.getParameterTypes();
		for (int i = 0; i < params.length; i++) {
			buf.append(toLongHand(params[i]));
			if (i+1 < params.length) {
				buf.append(",");
			}
		}
		buf.append(")");
		return buf.toString();
	}
	
	public static String toShortHand (IMethodBinding binding) {
		StringBuffer buf = new StringBuffer();
		String declClass = BindingUtil.toShortHand(binding.getDeclaringClass());
		if (! declClass.equals("UNKNOWN")) {
			buf.append(declClass);
			buf.append(".");
		}
		buf.append(binding.getName());
		buf.append("(");
		ITypeBinding[] params = binding.getParameterTypes();
		for (int i = 0; i < params.length; i++) {
			buf.append(params[i].getName());
			if (i+1 < params.length) {
				buf.append(",");
			}
		}
		buf.append(")");
		return buf.toString();
	}

	public static String toShortHand (IVariableBinding binding) {
		StringBuffer buf = new StringBuffer();
		String type = BindingUtil.toShortHand (binding.getDeclaringClass());
		if (! type.equals("UNKNOWN")) {
			buf.append(type);
			buf.append(".");
		}
		String ftype = BindingUtil.toShortHand(binding.getType());
		if (! type.equals("UNKNOWN")) {
			buf.append(ftype);
			buf.append(" ");
		}
		buf.append(binding.getName());
		return buf.toString();
	}

	public static String toShortHand (ITypeBinding binding) {
		if (binding == null) {
			return "UNKNOWN";
		}
		return binding.getName();
	}

	public static boolean isSubTypeOf (ITypeBinding sub, ITypeBinding sup) {
		if (sub == null || sup == null) {
			return false;
		}
		ITypeBinding superClass = sub.getSuperclass();
		if (BindingUtil.areEqual(superClass, sup)) {
			return true;
		}
		
		if (superClass != null) {
			if (isSubTypeOf (superClass, sup)) {
				return true;
			}
		}
			
		ITypeBinding[] interfaces = sub.getInterfaces();
		for (int i = 0; i < interfaces.length; i++) {
			ITypeBinding interfazz = interfaces[i];
			if (BindingUtil.areEqual(sup, interfazz)) {
				return true;
			}
			if (interfazz != null) {
				if (isSubTypeOf (interfazz, sup)) {
					return true;
				}
			}
		}
		
		return false;
	}
	
	public static Set<ITypeBinding> getCommonSuperTypes (ITypeBinding t1, ITypeBinding t2) {
		
		Set<ITypeBinding> superTypes = new HashSet<ITypeBinding>();
		
		ITypeBinding t1SuperClass = t1.getSuperclass();
		ITypeBinding t2SuperClass = t2.getSuperclass();
		
		// t1SuperClass is null unless t1 is a class binding 
		if (t1SuperClass != null && areEqual(t1SuperClass, t2SuperClass)) {
			superTypes.add(t1SuperClass);
		}
		
		ITypeBinding[] t1Interfaces = t1.getInterfaces();
		ITypeBinding[] t2Interfaces = t2.getInterfaces();
		
		superTypes.addAll(SetUtil.intersection(t1Interfaces, t2Interfaces));
		
		return superTypes;
	}
	
	public static ITypeBinding getTopLevelType (ITypeBinding t) {
		if (t==null) {
			return t;
		} else if (t.isMember()) {
			return null;
		} else if (t.isTopLevel()) {
			return t;
		} else {
			return t.getDeclaringClass();
		}
	}
	
//	public static IVariableBinding retrieveKey (IVariableBinding key, Map<IVariableBinding, ?> map) {
//		IVariableBinding mapKey = null;
//		for (IVariableBinding k : map.keySet()) {
//			if (BindingUtil.areEqual (key, k)) {
//				mapKey = k;
//			}
//		}
//		return mapKey;
//	}
	
//	public static boolean isVisible (ITypeBinding referenced, IMethodBinding fromMethodScope) 
//	{
//		// straightforward cases
//		if (referenced.isTopLevel()) {
//			int refModifiers = referenced.getModifiers();
//			if (Modifier.isPublic(refModifiers)) {
//				return true;
//			} else if (Modifier.isProtected(refModifiers)) {
//				// TODO check extension
//				// TODO check package
//			} else if (! Modifier.isPrivate(refModifiers)) {
//				// TODO check package
//			} else {
//				throw new RuntimeException();
//			}
//		}
//		else if (referenced.isMember()) {
//			IMethodBinding refMethod = referenced.getDeclaringMethod();
//			return areEqual(fromMethodScope, refMethod); 
//		}
//		
//		return false;
//	}
	
//	public static int hashCode (IBinding binding) {
//		switch (binding.getKind()){
//		case IBinding.METHOD:
//			return hashCode ((IMethodBinding) binding);
//		case IBinding.TYPE:
//			return hashCode ((ITypeBinding) binding);
//		case IBinding.VARIABLE:
//			return hashCode ((IVariableBinding) binding);
//		default:
//			return binding.hashCode();
//		}
//	}
//	
//	public static int hashCode (IMethodBinding binding) {
//		int hashCode = hashCode (binding.getDeclaringClass());
//		hashCode *= hashCode (binding.getParameterTypes());
//		hashCode *= binding.getName().hashCode();
//		return hashCode;
//	}
//	
//	public static int hashCode (IVariableBinding binding) {
//		int hashCode = binding.getName().hashCode();
//		hashCode *= binding.getDeclaringClass() != null ? 
//					 hashCode (binding.getDeclaringClass()) : hashCode (binding.getDeclaringMethod());
//		hashCode *= hashCode(binding.getType());
//		return hashCode;
//	}
//	
//	public static int hashCode (ITypeBinding binding) {
//		return binding.getQualifiedName().hashCode();
//	}
	
//	public static int hashCode (ITypeBinding[] binding) {
//		int hashCode = 1;
//		for (int i = 0; i < binding.length; i++) {
//			hashCode *= hashCode (binding[i]);
//		}
//		return hashCode;
//	}
	
//	public static int firstIndexOf (IMethodBinding binding, List<IMethodBinding> bindings) {
//		for (int i = 0; i < bindings.size(); i++) {
//			if (areEqual (binding, bindings.get(i))) {
//				return i;
//			}
//		}
//		
//		return -1;
//	}
}
