/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.util;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import ca.mcgill.cs.swevo.util.filter.IEqualityFilter;
import ca.mcgill.cs.swevo.util.filter.IFilter;
import ca.mcgill.cs.swevo.util.op.IConverter;

public class CollectionMap <T, F> {

	private Map<T, Collection<F>> map;
	
	public CollectionMap () {
		map = new HashMap<T, Collection<F>> ();
	}
	
	public CollectionMap (CollectionMap<T, F> other) {
		this();
		addAll (other);
	}
	
	public void add (T t, F f) {
		if (! map.containsKey(t)) {
			map.put(t, new ArrayList<F>());
		}
		
		map.get(t).add(f);
	}

	/**
	 * Adds the element f to all keys matching equalityFilter. 
	 * If none match, then f is added to t (which is supposed 
	 * to be equivalent to the equalityFilter)
	 */
	public void add (IEqualityFilter<T> equalityFilter, F f) {
		Collection<T> mapped = map(equalityFilter);
		
		if (mapped.isEmpty()) {
			T comparison = equalityFilter.getComparison();
			map.put(comparison, new ArrayList<F>());
			mapped = Collections.singleton(comparison);
		}
		
		for (T el : mapped) {
			map.get(el).add(f);
		}
	}
	
	private Set<T> map (IFilter<T> equalityFilter) {
		Set<T> mapped = new HashSet<T> ();
		for (T key : map.keySet()) {
			if (equalityFilter.passesFilter(key)) {
				mapped.add(key);
			}
		}
		return mapped;
	}
	
	public void addAll (T t, F...fs) {
		if (! map.containsKey(t)) {
			map.put(t, new ArrayList<F>());
		}
		
		for (F f : fs) {
			map.get(t).add(f);
		}
	}
	
	public void addAll (T t, Collection<? extends F> fCol) {
		if (! map.containsKey(t)) {
			map.put(t, new ArrayList<F>());
		}
		
		map.get(t).addAll(fCol);
	}
	
	public void addAll (CollectionMap<T, ? extends F> map) {
		for (T key : map.keySet()) {
			addAll (key, map.get(key));
		}
	}
	
	public Set<T> keySet () {
		return Collections.unmodifiableSet(map.keySet());
	}
	
	/**
	 * Returns the set of all key pairs. 
	 * Note, if a pair (A, B) is in the set, 
	 * we do not have (B, A) in the set. 
	 */
	@SuppressWarnings("unchecked")
	public LRTupleSet<T> keyPairSet () {
		LRTupleSet<T> pairs = new LRTupleSet<T>();
		Object[] keys = keySet().toArray();
		for (int i = 0; i < keys.length; i++) {
			for (int j = i + 1; j < keys.length; j++) {
				pairs.add(new LRTuple<T>((T) keys[i], (T) keys[j]));
			}
		}
		return pairs;
	}
	
	public Collection<F> get (T key) {
		if (! map.containsKey(key)) {
			map.put(key, new ArrayList<F>());
		}
		return map.get(key);
	}
	
	public Collection<F> get(IEqualityFilter<T> equalityFilter) {
		if (! containsKey(equalityFilter)) {
			map.put(equalityFilter.getComparison(), new ArrayList<F>());
		}
		return map.get(equalityFilter);
	}
	
	/**
	 * Returns all values associated with the given key equality filter.
	 * If no keys match the filter, then we return an empty collection
	 */
	public Collection<F> getAllValues (IEqualityFilter<T> equalityFilter) {
		Collection<F> col = new LinkedList<F>();
		for (T key : map.keySet()) {
			if (equalityFilter.passesFilter(key)) {
				col.addAll(map.get(key));
			}
		}
		return col;
	}
	
	public Set<F> excludeKey (T key) {
		Set<T> keys = new HashSet<T>(map.keySet());
		keys.remove(key);
		return subMap (keys).values();
	}
	
	/**
	 * Returns a the portion of this map that is defined by the input keys.
	 */
	public CollectionMap<T, F> subMap (Collection<T> keys) {
		CollectionMap<T, F> subMap = new CollectionMap<T, F>();
		for (T key : keys) {
			subMap.addAll(key, get(key));
		}
		return subMap;
	}
	
	/**
	 * @return All the set of values we find in the map's key collections
	 */
	public Set<F> values () {
		return new HashSet<F>(_values());
	}
	
	private Collection<F> _values () {
		Collection<F> values = new LinkedList<F>();
		for (T key : keySet()) {
			values.addAll(get(key));
		}
		
		return values;
	}
	
	/**
	 * Removes the given key
	 */
	public void removeKey (T key) {
		map.remove(key);
	}
	
	/**
	 * Removes all keys matching the current filter
	 */
	public void removeKey (IEqualityFilter<T> keyEq) {
		CollectionUtil.removeAll(map.keySet(), keyEq);
	}
		
	/**
	 * Removes all instances of this value from all keys
	 */
	public void removeValue (F value) {
		for (T key : map.keySet()) {
			get(key).remove(value);
		}
	}
	
	/**
	 * Removes all instances of the given value from this key
	 */
	public void removeFromKey (T key, F value) {
		get(key).removeAll(Collections.singleton(value)); 
	}
	
	/**
	 * Removes all values from the given key that match this filter 
	 */
	public void removeFromKey (T key, IFilter<F> values) {
		Collection<F> vals = get(key);
		CollectionUtil.removeAll(vals, values);
	}
	
	/**
	 * Clears the map
	 */
	public void clear () {
		map.clear();
	}
	
	@Override
	public String toString () {
		return map.toString();
	}
	
	/**
	 * For the given key, retains all values that pass the filter
	 */
	public void retainAllValues (T key, IFilter<F> filter) {
		Collection<F> col = map.get(key);
		col.retainAll(CollectionUtil.filterAndCopy(col, filter));
	}
	
	/**
	 * For each key, retains all values that pass the filter.
	 */
	public void retainAllValues (IFilter<F> filter) {
		for (T key : map.keySet()) {
			retainAllValues(key, filter);
		}
	}
	
	/**
	 * Applies the given transformation to each key. Otherwise, the map 
	 * is preserved.
	 */
	public void applyToKeys (IConverter<T, T> converter) {
		HashMap<T, Collection<F>> map = new HashMap<T, Collection<F>>();
		for (T key : keySet()) {
			Collection<F> values = this.map.get(key);
			map.put(converter.convert(key), values);
		}
		clear();
		for (T key : map.keySet()) {
			addAll(key, map.get(key));
		}
	}
	
	/**
	 * Performs a transpose operation and returns the result. 
	 * @return This map, but with all values mapped to the keys that hold them
	 */
	public CollectionMap<F, T> transpose () {
		CollectionMap<F, T> transpose = new CollectionMap<F, T>();
		for (T key : map.keySet()) {
			for (F val : get(key)) {
				transpose.add(val, key);
			}
		}
		return transpose;
	}
	
	/**
	 * Convenience method. Just like a regular map.
	 */
	public boolean containsKey (T key) {
		return map.containsKey(key);
	}
	
	/**
	 * Checks whether this map contains at least one key which matches the given 
	 * filter.
	 * @param equalityFilter A filter to match on keys
	 */
	public boolean containsKey (IFilter<T> equalityFilter) {
		boolean contains = false;
		for (T key : map.keySet()) {
			contains = contains ? true : equalityFilter.passesFilter(key); 
		}
		return contains;
	}
}
